# -*- coding: utf-8 -*-
"""
@author: walroth
"""

# Standard library imports
import sys
import traceback

# Other imports
import numpy as np
from matplotlib import pyplot as plt

# Qt imports
import pyqtgraph as pg
from pyqtgraph import Qt
from pyqtgraph.Qt import QtWidgets

from xdart.utils import catch_h5py_file as catch

from ...widgets import MaskWidget

# from icecream import ic; ic.configureOutput(prefix='', includeContext=True)

PUSHBUTTON_STYLESHEET = """
            QPushButton {
              background-color: #505F69;
              border: 1px solid #32414B;
              color: #F0F0F0;
              border-radius: 4px;
              padding: 3px;
              outline: none;
              min-width: 20px;
            }
            QPushButton:disabled {
              background-color: #32414B;
              border: 1px solid #32414B;
              color: #787878;
              border-radius: 4px;
              padding: 3px;
            }
            
            QPushButton:checked {
              background-color: #32414B;
              border: 1px solid #32414B;
              border-radius: 4px;
              padding: 3px;
              outline: none;
            }
            
            QPushButton:checked:disabled {
              background-color: #19232D;
              border: 1px solid #32414B;
              color: #787878;
              border-radius: 4px;
              padding: 3px;
              outline: none;
            }
            
            QPushButton:checked:selected {
              background: #1464A0;
              color: #32414B;
            }
            
            QPushButton::menu-indicator {
              subcontrol-origin: padding;
              subcontrol-position: bottom right;
              bottom: 4px;
            }
            
            QPushButton:pressed {
              background-color: #19232D;
              border: 1px solid #19232D;
            }
            
            QPushButton:pressed:hover {
              border: 1px solid #148CD2;
            }
            
            QPushButton:hover {
              border: 1px solid #148CD2;
              color: #F0F0F0;
            }
            
            QPushButton:selected {
              background: #1464A0;
              color: #32414B;
            }
            
            QPushButton:hover {
              border: 1px solid #148CD2;
              color: #F0F0F0;
            }
            
            QPushButton:focus {
              border: 1px solid #1464A0;
            }
        """
QFileDialog = QtWidgets.QFileDialog

# This module imports
from .displayFrameUI import Ui_Form
from ...gui_utils import RectViewBox, get_rect
from ...widgets import XDImageWidget
import xdart.utils as ut

formats = [
    str(f.data(), encoding='utf-8').lower() for f in
    Qt.QtGui.QImageReader.supportedImageFormats()
]

class displayFrameWidget(Qt.QtWidgets.QWidget):
    """Widget for displaying 2D image data and 1D plots from EwaldSphere
    objects. 
    
    attributes:
        auto_last: bool, whether to automatically select latest arch
        curve1: pyqtgraph pen, overall data line
        curve2: pyqtgraph pen, individual arch data line
        histogram: pyqtgraph HistogramLUTWidget, used for adjusting min
            and max level for image
        image: pyqtgraph ImageItem, displays the 2D data
        image_plot: pyqtgraph plot, for 2D data
        image_win: pyqtgraph GraphicsLayoutWidget, layout for the 2D
            data
        imageViewBox: RectViewBox, used to set behavior of the image
            plot
        plot: pyqtgraph plot, for 1D data
        plot_layout: QVBoxLayout, for holding the 1D plotting widgets
        plot_win: pyqtgraph GraphicsLayoutWidget, layout for the 1D
            data
        sphere: EwaldSphere, unused.
        ui: Ui_Form from qtdeisgner
    
    methods:
        get_arch_data_2d: Gets 2D data from an arch object
        get_sphere_data_2d: Gets overall 2D data for the sphere
        update: Updates the displayed image and plot
        update_image: Updates image data based on selections
        update_plot: Updates plot data based on selections
    """
    def __init__(self, sphere, arch, parent=None):
        # ic()
        _translate = Qt.QtCore.QCoreApplication.translate
        super().__init__(parent)
        self.ui = Ui_Form()
        self.ui.setupUi(self)
        self.ui.imageUnit.setItemText(0, _translate("Form", "2" + u"\u03B8"))
        self.ui.plotUnit.setItemText(0, _translate("Form", "2" + u"\u03B8"))

        # Data object initialization
        self.sphere = sphere
        self.arch = arch

        # State variable initialization
        self.auto_last = False

        # Image pane setup
        self.image_layout = Qt.QtWidgets.QHBoxLayout(self.ui.imageFrame)
        self.image_layout.setContentsMargins(0, 0, 0, 0)
        self.image_layout.setSpacing(0)
        self.image_widget = XDImageWidget()
        self.image_layout.addWidget(self.image_widget)

        # Image pane signal connections
        self.ui.imageMethod.setCurrentIndex(1)
        self.ui.imageMethod.setEnabled(False)

        self.plot_layout = Qt.QtWidgets.QVBoxLayout(self.ui.plotFrame)
        self.plot_layout.setContentsMargins(0, 0, 0, 0)
        self.plot_win = pg.GraphicsLayoutWidget()
        self.plot_layout.addWidget(self.plot_win)
        vb = RectViewBox()
        self.plot = self.plot_win.addPlot(viewBox=vb)
        self.curve1 = self.plot.plot(pen=(50,100,255))
        self.curve2 = self.plot.plot(
            pen=(200,50,50,200), 
            symbolBrush=(200,50,50,200), 
            symbolPen=(0,0,0,0), 
            symbolSize=4
        )

        self.ui.plotMethod.setCurrentIndex(1)
        self.ui.plotMethod.setEnabled(False)
        
        # Signal connections
        self.ui.imageIntRaw.activated.connect(self.update_image)
        self.ui.imageMethod.activated.connect(self.update_image)
        self.ui.imageUnit.activated.connect(self.update_image)
        self.ui.imageNRP.activated.connect(self.update_image)
        self.ui.imageMask.stateChanged.connect(self.update_image)
        self.ui.shareAxis.stateChanged.connect(self.update)
        self.ui.plotMethod.activated.connect(self.update_plot)
        self.ui.plotUnit.activated.connect(self.update_plot)
        self.ui.plotNRP.activated.connect(self.update_plot)
        self.ui.plotOverlay.stateChanged.connect(self.update_plot)

        # Fix width of buttons
        self.ui.pushRight.setStyleSheet(PUSHBUTTON_STYLESHEET)
        self.ui.pushRightLast.setStyleSheet(PUSHBUTTON_STYLESHEET)
        self.ui.pushLeft.setStyleSheet(PUSHBUTTON_STYLESHEET)
        self.ui.pushLeftLast.setStyleSheet(PUSHBUTTON_STYLESHEET)

        self.mask_widget = MaskWidget()
        self.mask_widget.hide()
        self.ui.setMaskButton.clicked.connect(self.launch_mask_widget)
        self.mask_widget.requestMask.connect(self._send_to_mask_widget)
        self.mask_widget.newMask.connect(self._set_new_mask)

        #self.update()

    def update(self):
        """Updates image and plot frames based on toolbar options
        """
        # ic()
        # Sets title text
        if self.arch.idx is None:
            self.ui.labelCurrent.setText(self.sphere.name)
        else:
            self.ui.labelCurrent.setText("Image " + str(self.arch.idx))

        if self.ui.shareAxis.isChecked():
            self.ui.plotUnit.setCurrentIndex(self.ui.imageUnit.currentIndex())
            self.ui.plotUnit.setEnabled(False)
            self.plot.setXLink(self.image_widget.image_plot)
        
        else:
            self.plot.setXLink(None)
            self.ui.plotUnit.setEnabled(True)
        
        try:
            self.update_image()
        except TypeError:
            return False
        try:
            self.update_plot()
        except TypeError:
            return False
        return True
    
    def update_image(self):
        """Updates image plotted in image frame
        """
        # ic()
        if self.sphere.name == 'null_main':
            data = np.arange(100).reshape(10,10)
            rect = Qt.QtCore.QRect(1,1,1,1)
        else:
            try:
                if self.arch.idx is not None:
                    for i in range(10):
                        try:
                            data, rect = self.get_arch_data_2d()
                            break
                        except AttributeError:
                            pass

                else:
                    data, rect = self.get_sphere_data_2d()
            except (TypeError, IndexError):
                data = np.arange(100).reshape(10, 10)
                rect = Qt.QtCore.QRect(1, 1, 1, 1)
        try:
            self.image_widget.setImage(data)
            self.image_widget.setRect(rect)
        except UnboundLocalError:
            data = np.arange(100).resize((10, 10))
        
        return data

    def get_arch_data_2d(self):
        """Returns data and QRect for data in arch
        """
        # ic()
        with self.arch.arch_lock:
            int_data = self.arch.int_2d
        
        if self.ui.imageIntRaw.currentIndex() == 0:
            data, corners, sigma = read_NRP(self.ui.imageNRP, int_data)
            
            rect = get_rect(
                get_xdata(self.ui.imageUnit, int_data)[corners[2]:corners[3]], 
                int_data.chi[corners[0]:corners[1]]
            )
        
        elif self.ui.imageIntRaw.currentIndex() == 1:
            with self.arch.arch_lock:
                if self.ui.imageNRP.currentIndex() == 0:
                    if self.arch.map_norm is None or self.arch.map_norm == 0:
                        data = self.arch.map_raw.copy()
                    else:
                        data = self.arch.map_raw.copy()/self.arch.map_norm
                else:
                    data = self.arch.map_raw.copy()
                if self.ui.imageMask.isChecked():
                    data.ravel()[self.arch.mask] = data.max()
                    if self.sphere.global_mask is not None:
                        data.ravel()[self.sphere.global_mask] = data.max()
            rect = get_rect(
                np.arange(data.shape[0]), 
                np.arange(data.shape[1]),
            )
        
        return data, rect

    def get_sphere_data_2d(self):
        """Returns data and QRect for data in sphere
        """
        # ic()
        with self.sphere.sphere_lock:
            if self.ui.imageMethod.currentIndex() == 0:
                int_data = self.sphere.mgi_2d
                if type(int_data.ttheta) == int:
                    self.ui.imageMethod.setCurrentIndex(1)
                    int_data = self.sphere.bai_2d
            elif self.ui.imageMethod.currentIndex() == 1:
                int_data = self.sphere.bai_2d

        data, corners, sigma = read_NRP(self.ui.imageNRP, int_data)

        rect = get_rect(
            get_xdata(self.ui.imageUnit, int_data)[corners[2]:corners[3]], 
            int_data.chi[corners[0]:corners[1]]
        )

        return data, rect
    
    def update_plot(self):
        """Updates data in plot frame
        """
        # ic()
        if self.sphere.name == 'null_main':
            data = (np.arange(100), np.arange(100))
            self.curve1.setData(data[0], data[1])
            self.curve2.setData(data[0], data[1])
            return data
        
        else:
            try:
                with self.sphere.sphere_lock:
                    if self.ui.plotMethod.currentIndex() == 0:
                        sphere_int_data = self.sphere.mgi_1d
                        if type(sphere_int_data.ttheta) == int:
                            self.ui.plotMethod.setCurrentIndex(1)
                            sphere_int_data = self.sphere.bai_1d
                    elif self.ui.plotMethod.currentIndex() == 1:
                        sphere_int_data = self.sphere.bai_1d

                s_ydata, corners, s_sigma = read_NRP(self.ui.plotNRP, sphere_int_data)
                s_xdata = get_xdata(self.ui.plotUnit, sphere_int_data)[corners[0]:corners[1]]

                if self.arch.idx is not None:
                    with self.arch.arch_lock:
                        arc_int_data = self.arch.int_1d

                    if self.ui.plotOverlay.isChecked():
                        self.curve1.setData(s_xdata, s_ydata)
                    else:
                        self.curve1.clear()

                    a_ydata, corners, a_sigma = read_NRP(self.ui.plotNRP, arc_int_data)
                    a_xdata = get_xdata(self.ui.plotUnit, arc_int_data)[corners[0]:corners[1]]
                    self.curve2.setData(a_xdata, a_ydata)

                    return a_xdata, a_ydata, a_sigma

                else:
                    self.curve1.setData(s_xdata, s_ydata)
                    self.curve2.clear()

                    return s_xdata, s_ydata, s_sigma

            except (TypeError, IndexError):
                data = (np.arange(100), np.arange(100))
                self.curve1.setData(data[0], data[1])
                self.curve2.setData(data[0], data[1])
                return data

    def save_image(self):
        """Saves currently displayed image. Formats are automatically
        grabbed from Qt. Also implements tiff saving.
        """
        # ic()
        ext_filter = "Images ("
        for f in formats:
            ext_filter += "*." + f + " "

        ext_filter += "*.tiff)"

        fname, _ = QFileDialog.getSaveFileName(filter=ext_filter)
        if fname == '':
            return

        _, ext = fname.split('.')
        if ext.lower() in formats:
            self.image_widget.imageItem.save(fname)
        
        elif ext.lower() == 'tiff':
            data = self.update_image()
            plt.imsave(fname, data.T, cmap='viridis')
    
    def save_array(self):
        """Saves currently displayed data. Currently supports .xye
        and .csv.
        """
        # ic()
        fname, _ = QFileDialog.getSaveFileName(
            filter="XRD Files (*.xye *.csv)"
        )
        if fname == '':
            return

        xdata, ydata, sigma = self.update_plot()

        _, ext = fname.split('.')
        if ext.lower() == 'xye':
            ut.write_xye(fname, xdata, ydata, sigma)
        
        elif ext.lower() == 'csv':
            ut.write_csv(fname, xdata, ydata, sigma)

    def launch_mask_widget(self, q=None):
        self.update_mask_widget()
        if self.arch.idx is None:
            with self.sphere.file_lock:
                try:
                    _arch = self.sphere.arches[0]
                    data = _arch.map_raw.copy()
                except KeyError:
                    traceback.print_exc()
                    return
            mask = np.zeros_like(data)
            if self.sphere.global_mask is not None:
                mask.ravel()[self.sphere.global_mask] = 1
            self.mask_widget.set_data(data, base=mask)
            self.mask_widget.blockSignals(True)
            self.mask_widget.ui.archList.setCurrentIndex(0)
            self.mask_widget.blockSignals(False)
        else:
            self.mask_widget.set_data(self.arch.map_raw, base=self.arch.get_mask())
            idx = self.mask_widget.ui.archList.findText(str(self.arch.idx))
            if idx < 0:
                idx = 0
            self.mask_widget.blockSignals(True)
            self.mask_widget.ui.archList.setCurrentIndex(idx)
            self.mask_widget.blockSignals(False)
        self.mask_widget.show()

    def update_mask_widget(self):
        self.mask_widget.blockSignals(True)
        try:
            idx = self.mask_widget.ui.archList.currentIndex()
            idx = max(idx, 0)
            self.mask_widget.ui.archList.clear()
            i = 0
            self.mask_widget.ui.archList.insertItem(i, "Global")
            i += 1
            for idx in self.sphere.arches.index:
                self.mask_widget.ui.archList.insertItem(i, str(idx))
                i += 1
            idx = min(idx, self.mask_widget.ui.archList.count() - 1)
            self.mask_widget.ui.archList.setCurrentIndex(idx)
        except:
            traceback.print_exc()
            t, v, tb = sys.exc_info()
            self.mask_widget.blockSignals(False)
            raise(t, v, tb)
        finally:
            self.mask_widget.blockSignals(False)

    def _send_to_mask_widget(self, q):
        if q >= 0:
            with self.sphere.file_lock:
                _arch = self.sphere.arches[q]
            self.mask_widget.set_data(_arch.map_raw, base=_arch.get_mask())
        elif self.sphere.global_mask is not None:
            base = np.zeros_like(self.mask_widget.data)
            base.ravel()[self.sphere.global_mask] = 1
            self.mask_widget.set_data(self.mask_widget.data, base=base)
        else:
            self.mask_widget.clear_mask()

    def _set_new_mask(self, idx, mask):
        mask_ids = np.arange(self.mask_widget.data.size)[mask.ravel() == 1]
        if idx < 0:
            self.sphere.global_mask = mask_ids
            with self.sphere.file_lock:
                self.sphere.save_to_h5()
        else:
            if self.arch.idx is not None and self.arch.idx == idx:
                _arch = self.arch
            else:
                try:
                    with self.sphere.file_lock:
                        _arch = self.sphere.arches[idx]
                except KeyError:
                    traceback.print_exc()
                    return
            _arch.mask = mask_ids
            with self.sphere.file_lock:
                with catch(self.sphere.data_file, 'a') as file:
                    arches = file['arches']
                    _arch.save_to_h5(arches)


def read_NRP(box, int_data):
    """Reads the norm, raw, pcount option box and returns
    appropriate ydata.
    
    args:
        box: QComboBox, list of choices for picking data to return
        int_data: int_nd_data object, data to parse
    
    returns:
        data: numpy array, non-zero region from nzarray based on
            choices in box
        corners: tuple, the bounds of the non-zero region of the
            dataset
    """
    # ic()
    if box.currentIndex() == 0:
        nzarr = int_data.norm
        sigmanz = int_data.sigma
    elif box.currentIndex() == 1:
        nzarr = int_data.raw
        sigmanz = int_data.sigma_raw
    elif box.currentIndex() == 2:
        nzarr = int_data.pcount
        sigmanz = None
    data = nzarr.data[()].T
    if sigmanz is not None:
        sigma = sigmanz.data[()].T
    else:
        sigma = np.zeros_like(data)
    corners = nzarr.corners
    
    if data.size == 0:
        data = np.zeros(int_data.norm.shape)
        sigma = np.zeros_like(data)
        if len(corners) == 2:
            corners = [0, int_data.norm.shape[0]]
        elif len(corners) == 4:
            corners = [0, int_data.norm.shape[0], 0, int_data.norm.shape[1]]
    return data, corners, sigma


def get_xdata(box, int_data):
    """Reads the unit box and returns appropriate xdata
    
    args:
        box: QComboBox, list of options
        int_data: int_nd_data object, data to parse
    
    returns:
        xdata: numpy array, x axis data for plot.
    """
    # ic()
    if box.currentIndex() == 0:
        xdata = int_data.ttheta
    elif box.currentIndex() == 1:
        xdata = int_data.q
    return xdata



# -*- coding: utf-8 -*-
"""
Created on Wed Aug 28 16:39:57 2019

@author: walroth
"""
import os
from collections import OrderedDict
import pandas as pd
from copy import deepcopy

from xdart.utils import soft_list_eval
from ..operation import Operation



inputs = OrderedDict(
        spec_file_path=None,
        spec_file_name=None
        )

outputs = OrderedDict(
        header=dict(
                meta={},
                motors={},
                motors_r={},
                detectors={},
                detectors_r={}
                ),
        scans={},
        scans_meta={},
        last_line_read=dict(
                number=0,
                text=''
                ),
        current_scan={
            'num': 0,
            'data': None,
            'meta': None
        }
    )

class LoadSpecFile(Operation):
    """Operation for loading in data from a spec file. Currently
    unused, will be expanded on later for future analysis toolkit.
    """
    date_format = '%a %b %d %H:%M:%S %Y'

    def __init__(self):
        super(LoadSpecFile, self).__init__(inputs, outputs)
    
    def run(self):
        full_path = os.path.join(self.inputs['spec_file_path'],
                                 self.inputs['spec_file_name'])
        with open(full_path, 'r') as file:
            self._read_spec(file)
        
        return self.outputs
    
    
    def _read_spec(self, file):
        # iterate lines and assign to either head or scan lists
        state = 'beginning'
        head = []

        for lin_num, lin in enumerate(file):
            line = lin.split()
            self.outputs['last_line_read']['number'] = lin_num
            self.outputs['last_line_read']['text'] = lin
            # blank lines are used as breaks separating scans
            if line == []:
                # state flag used to control how lines are parsed
                if state == 'beginning':
                    continue

                elif state == 'head':
                    self._parse_header(head)
                    head = []

                elif state == 'scan':
                    continue

            else:
                # first item defines what type of info it is
                key = line[0]
                if '#F' in key:
                    state = 'head'
                elif '#S' in key:
                    state = 'scan'
                    scan_num = int(line[1])

                if state == 'head':
                    head.append(line)

                elif state == 'scan':
                    self._parse_scan(line, scan_num)
    
    
    def _parse_header(self, head):
        if head == []:
            return None
        
        else:
            meta = {}
            motors = {}
            motors_r = {}
            detectors = {}
            detectors_r = {}
            for line in head:
                key = line[0][1:]
                if key == 'F': 
                    meta['File'] = line[1:]

                elif key == 'E':
                    meta['Epoch'] = eval(line[1])

                elif key == 'D':
                    meta['Date'] = ' '.join(line[1:])

                elif key == 'C':
                    meta['Comment'] = ' '.join(line[1:])
                    for i, val in enumerate(line):
                        if val == 'User' and line[i + 1] == '=':
                            meta['User'] = line[i+2]
                
                elif 'O' in key:
                    num = int(key[1:])
                    motors[num] = line[1:]
                
                elif 'o' in key:
                    num = int(key[1:])
                    motors_r[num] = line[1:]
                
                elif 'J' in key:
                    num = int(key[1:])
                    detectors[num] = line[1:]
                
                elif 'j' in key:
                    num = int(key[1:])
                    detectors_r[num] = line[1:]
            
            self.outputs['header']['meta'].update(meta)
            self.outputs['header']['motors'].update(motors)
            self.outputs['header']['motors_r'].update(motors_r)
            self.outputs['header']['detectors'].update(detectors)
            self.outputs['header']['detectors_r'].update(detectors_r)
    
    def _parse_scan(self, line, scan_num):
        if line == []:
            return None

        flag = line[0]
        if '#' in flag:
            if 'S' in flag:
                self.outputs['scans_meta'][scan_num] = \
                    {'Goniometer': {}, 'Motors': {}}
                self.outputs['scans_meta'][scan_num]['Command'] = \
                    ' '.join(line[2:])
                self.outputs['scans_meta'][scan_num]['Type'] = line[2]

            elif 'D' in flag:
                self.outputs['scans_meta'][scan_num]['Date'] = \
                    ' '.join(line[1:])

            elif 'T' in flag or 'M' in flag:
                self.outputs['scans_meta'][scan_num]['Counter'] = \
                    {'Amount': eval(line[1]), 'Type': line[2]}

            elif 'G' in flag:
                key = int(flag[2:])
                self.outputs['scans_meta'][scan_num]['Goniometer'][key] = \
                    soft_list_eval(line[1:])

            elif 'Q' in flag:
                self.outputs['scans_meta'][scan_num]['HKL'] = \
                    soft_list_eval(line[1:])

            elif 'P' in flag:
                motor_num = int(flag[2:])
                names = self.outputs['header']['motors'][motor_num]
                positions = [eval(x) for x in line[1:]]
                self.outputs['scans_meta'][scan_num]['Motors'].update(
                    {name: position for name, position in 
                    zip(names, positions)}
                )
            # TODO: decide what to do with N;

            elif 'L' in flag:
                self.outputs['scans'][scan_num] = pd.DataFrame(columns=line[1:])

        else:
            vals = soft_list_eval(line)
            try:
                idx = self.outputs['scans'][scan_num].index[-1] + 1
                self.outputs['scans'][scan_num].loc[idx] = vals
            except IndexError:
                self.outputs['scans'][scan_num].loc[0] = vals
    


import copy

import numpy as np
from pyFAI import units
from scipy.interpolate import RectBivariateSpline

from .. import _utils as utils

# from icecream import ic; ic.configureOutput(prefix='', includeContext=True)


class int_1d_data_static:
    """Container for 1-dimensional integration data returned by pyFAI.

    attributes:
        norm: np.ndarray, integrated signal
        ttheta: numpy array, two-theta angle
        q: numpy array, q values

    methods:
        from_hdf5: Loads in data from an hdf5 file
        from_result: Parses data in from result returned by pyFAI
        parse_unit: Ensures both q and ttheta data is held
        to_hdf5: Saves data to hdf5 file
    """

    def __init__(self, norm=None, ttheta=0, q=0, i_qz=None, qz=0, i_qxy=0, qxy=0):
        """
        norm: np.ndarray, integrated signal
        ttheta: numpy array, two-theta angle
        q: numpy array, q values
        """
        self.norm = norm
        self.ttheta = ttheta
        self.q = q
        self.i_qz = i_qz
        self.qz = qz
        self.i_qxy = i_qxy
        self.qxy = qxy

    def from_result(self, result, wavelength, unit=None):
        """Parses out result obtained by pyFAI AzimuthalIntegrator.

        args:
            result: object returned by AzimuthalIntegrator
            wavelength: float, energy of the beam in meters
        """
        self.ttheta, self.q = self.parse_unit(
            result, wavelength, unit=unit)

        self.norm = result.intensity

    def parse_unit(self, result, wavelength, unit=None):
        """Helper function to take integrator result and return a two
        theta and q array regardless of the unit used for integration.

        args:
            result: result from 1dintegrator
            wavelength: wavelength for conversion in meters

        returns:
            int_1d_2theta: two theta array
            int_1d_q: q array
        """
        if wavelength is None:
            return result.radial, None

        if unit is None:
            unit = result.unit

        # if result.unit == units.TTH_DEG or str(result.unit) == '2th_deg':
        if unit == units.TTH_DEG or str(unit) == '2th_deg':
            int_1d_2theta = result.radial
            int_1d_q = (
                    (4 * np.pi / (wavelength * 1e10)) *
                    np.sin(np.radians(int_1d_2theta / 2))
            )
        # elif result.unit == units.Q_A or str(result.unit) == 'q_A^-1':
        elif unit == units.Q_A or str(unit) == 'q_A^-1':
            int_1d_q = result.radial
            int_1d_2theta = (
                2 * np.degrees(np.arcsin(
                    int_1d_q * (wavelength * 1e10) / (4 * np.pi)))
            )

        # TODO: implement other unit options for unit
        return np.asarray(int_1d_2theta, dtype=np.float32), np.asarray(int_1d_q, dtype=np.float32)

    def to_hdf5(self, grp, compression=None):
        """Saves data to hdf5 file.

        args:
            grp: h5py Group or File, where the data will be saved
            compression: str, compression algorithm to use. See h5py
                documentation.
        """
        keys = self.__dict__.keys()
        for key in keys:
            if key in grp:
                del grp[key]
        utils.attributes_to_h5(self, grp, keys, compression=compression)

    def from_hdf5(self, grp):
        """Loads in data from hdf5 file.

        args:
            grp: h5py Group or File, object to load data from.
        """
        keys = self.__dict__.keys()
        utils.h5_to_attributes(self, grp, keys)

    def __setattr__(self, name, value):
        """Ensures raw, norm, and pcount are nzarray1d objects.
        """
        self.__dict__[name] = np.asarray(value)

    def __add__(self, other):
        out = self.__class__()
        out.norm = self.norm + other.norm
        out.ttheta = copy.deepcopy(self.ttheta)
        out.q = copy.deepcopy(self.q)
        return out


class int_2d_data_static(int_1d_data_static):
    """Container for 2-dimensional integration data returned by pyFAI.

    attributes:
        norm: nzarray2d, integrated signal normalized by number of
            pixels
        ttheta: numpy array, two-theta angle
        q: numpy array, q values
        chi: numpy array, chi values

    methods:
        from_hdf5: Loads in data from an hdf5 file
        from_result: Parses data in from result returned by pyFAI
        parse_unit: Ensures both q and ttheta data is held
        to_hdf5: Saves data to hdf5 file
    """

    # def __init__(self, i_tthChi=np.zeros(0), i_qChi=np.zeros(0), ttheta=0, q=0, chi=0,
    def __init__(self, i_tthChi=0, i_qChi=0, ttheta=0, q=0, chi=0,
                 i_QxyQz=0, qz=0, qxy=0):
        """
        raw: nzarray2d, raw integrated signal
        pcount: nzarray2d, how many pixels in each bin
        norm: nzarray2d, integrated signal normalized by number of
            pixels
        ttheta: numpy array, two-theta angle
        q: numpy array, q values
        chi: numpy array, chi values
        """
        self.i_tthChi = i_tthChi
        self.i_qChi = i_qChi
        self.ttheta = ttheta
        self.q = q
        self.chi = chi
        self.i_QxyQz = i_QxyQz
        self.qz = qz
        self.qxy = qxy
        self.q_from_tth = True
        self.tth_from_q = True

    def from_result(self, result, wavelength, unit=None,
                    i_QxyQz=0, qz=0, qxy=0):
        """Parses out result obtained by pyFAI AzimuthalIntegrator.
        args:
            result: object returned by AzimuthalIntegrator
            wavelength: float, energy of the beam in meters
        """
        if unit is None:
            unit = result.unit

        self.parse_unit(result, wavelength, unit=unit)

        self.i_QxyQz = i_QxyQz
        self.qz = qz
        self.qxy = qxy

    def parse_unit(self, result, wavelength, unit=None):
        """Helper function to take integrator result and return a two
        theta and q array regardless of the unit used for integration.

        args:
            result: result from 1d integrator
            wavelength: wavelength for conversion in meters

        returns:
            int_1d_2theta: two theta array
            int_1d_q: q array
        """
        if wavelength is None:
            return result.radial, None

        if unit is None:
            unit = result.unit

        self.chi = chi = result.azimuthal

        if unit == units.TTH_DEG or str(unit) == '2th_deg':
            self.i_tthChi = result.intensity
            self.ttheta = tth = result.radial
            self.tth_from_q = False
            if self.q_from_tth:
                tth_range = np.asarray([tth[0], tth[-1]])
                q_range = (4 * np.pi / (wavelength * 1e10)) * np.sin(np.radians(tth_range / 2))
                qtth = (4 * np.pi / (wavelength * 1e10)) * np.sin(np.radians(tth / 2))
                self.q = q = np.linspace(q_range[0], q_range[1], len(tth))

                spline = RectBivariateSpline(chi, qtth, result.intensity)
                self.i_qChi = spline(chi, q)

        elif unit == units.Q_A or str(unit) == 'q_A^-1':
            self.i_qChi = result.intensity
            self.q = q = result.radial
            self.q_from_tth = False
            if self.tth_from_q:
                q_range = np.array([q[0], q[-1]])
                tth_range = 2 * np.degrees(np.arcsin(q_range * (wavelength * 1e10) / (4 * np.pi)))
                tthq = 2 * np.degrees(np.arcsin(q * (wavelength * 1e10) / (4 * np.pi)))
                self.ttheta = tth = np.linspace(tth_range[0], tth_range[1], len(q))

                spline = RectBivariateSpline(chi, tthq, result.intensity)
                self.i_tthChi = spline(chi, tth)

        # TODO: implement other unit options for unit

    def from_hdf5(self, grp):
        """Loads in data from hdf5 file.

        args:
            grp: h5py Group or File, object to load data from.
        """
        super().from_hdf5(grp)
        utils.h5_to_attributes(
            self, grp, ['chi', 'i_QxyQz', 'qz', 'qxy', 'q_from_tth', 'tth_from_q']
        )

    def to_hdf5(self, grp, compression=None):
        """Saves data to hdf5 file.

        args:
            grp: h5py Group or File, where the data will be saved
            compression: str, compression algorithm to use. See h5py
                documentation.
        """
        super().to_hdf5(grp, compression)
        utils.attributes_to_h5(
            self, grp, ['chi', 'i_QxyQz', 'qz', 'qxy', 'q_from_tth', 'tth_from_q'],
            compression=compression
        )

    def __setattr__(self, name, value):
        """Ensures all saved objects are np.ndarray objects.
        """
        self.__dict__[name] = np.asarray(value)

    def __add__(self, other):
        out = self.__class__()

        out.i_qChi = self.i_qChi + other.i_qChi
        out.i_tthChi = self.i_tthChi + other.i_tthChi
        out.i_QxyQz = self.i_QxyQz + other.i_QxyQz

        out.ttheta = copy.deepcopy(other.ttheta)
        out.q = copy.deepcopy(other.q)
        out.chi = copy.deepcopy(other.chi)
        out.qz = copy.deepcopy(other.qz)
        out.qxy = copy.deepcopy(other.qxy)

        return out

    def __sub__(self, other):
        out = self.__class__()

        out.i_qChi = self.i_qChi - other.i_qChi
        out.i_tthChi = self.i_tthChi - other.i_tthChi
        out.i_QxyQz = self.i_QxyQz - other.i_QxyQz

        out.ttheta = copy.deepcopy(other.ttheta)
        out.q = copy.deepcopy(other.q)
        out.chi = copy.deepcopy(other.chi)
        out.qz = copy.deepcopy(other.qz)
        out.qxy = copy.deepcopy(other.qxy)

        return out



import numpy as np
from scipy.optimize import curve_fit
import pylab

from bServer_funcs import *

# Useful information for calibration goes here
data_path = "E:/SLAC/Data/SSRL/BL21/Schelhas/Dec2018/images/"
calib_name = "b_stone_direct_beam_scan1_"
csv_path = "E:/SLAC/Data/SSRL/BL21/Schelhas/Dec2018/"
csv_name = "direct_beam_scan1.csv"
db_pixel = [253, 95]  # direct beam pixel
pix_size = 172.0  # pixel size in microns


def read_TIFF(file):
    print("Reading TIFF file here...")
    try:
        im = open(file, 'rb')
        im.seek(4096)   # skip the first 4096 bytes of header info for TIFF images
        arr = np.fromstring(im.read(), dtype='int32')
        im.close()
        arr.shape = (195, 487)
        #arr = np.fliplr(arr)  #for the way mounted at BL2-1
        print(np.shape(arr))
        print(len(arr))
        return arr
    except:
        print("Error reading file: %s" % file)
        return None


def read_RAW(file):
    print("Reading RAW file here...")
    try:
        im = open(file, 'rb')
        arr = np.fromstring(im.read(), dtype='int32')
        im.close()
        arr.shape = (195, 487)
        #arr = np.fliplr(arr)  #for the way mounted at BL2-1
        return arr
    except:
        print("Error reading file: %s" % file)
        return None

def csvread(filename):  
    print("Reading CSV file here...")
    csv = open(filename)
    line = csv.readline()
    temp = line.split(',')
    xi = 1
    yi = 4
    i0i = 3
    x = []
    y = []
    i0 = []
    line = csv.readline()
    while line:
        temp = line.split(",")
        x = np.append(x, float(temp[xi]))
        y = np.append(y, float(temp[yi]))       
        i0 = np.append(i0, float(temp[i0i]))
        line = csv.readline()
    csv.close()
    return x, y, i0


def gauss_linbkg(x, m, b, x0, intint, fwhm):
    return m*x + b + intint*(2./fwhm)*np.sqrt(np.log(2.)/np.pi)*np.exp(-4.*np.log(2.)*((x-x0)/fwhm)**2)

    
def Gauss_fit(x, y):
    pguess = [0, 0, np.argmax(y), np.max(y), 5.0]  # linear background (2), pos, intensity, fwhm
    popt, pcov = curve_fit(gauss_linbkg, x, y, p0=pguess)
    return popt

    
def simple_line(x, m, b):
    return m*x + b


# Read data points
scan_type = get_last_scan_details()['scan_type']
columns = [scan_type] + get_counter_mnemonics()
all_columns = columns + ['None' for _ in range(128-len(columns))]

data_points = get_plot_points(columns=all_columns)


# Read CSV file, get step size and number of points in scan
x, y, i0 = csvread(csv_path + csv_name)
num_points = len(x)
calib_tth_steps = abs(x[1] - x[0])
x = []
y = []
i0 = []

# Read images, take line cut, fit peak for Al2O3 calibration scan
pks = []
pylab.figure()
for i in range(0, num_points):
    print(i)
    filename = data_path + calib_name + str(i).zfill(4) + ".raw"
    data = read_RAW(filename)
    x = np.arange(0, np.shape(data)[1])
    y = data[db_pixel[1], :]
    y += data[db_pixel[1] + 1, :]
    y += data[db_pixel[1] - 1, :]
    y += data[db_pixel[1] + 2, :]
    y += data[db_pixel[1] - 2, :]
    popt = Gauss_fit(x, y)
    pylab.cla()
    pylab.plot(x,y, 'b.')
    pylab.plot(x, gauss_linbkg(x, *popt), 'r-')
    pks = np.append(pks, popt[2])

# Fit line to the extracted peak positions and determine the sample to detector distance
x = np.arange(num_points)
lin_fit, pcov = curve_fit(simple_line, pks, x*calib_tth_steps + 0.00)
det_R = 1.0/np.tan(abs(lin_fit[0])*np.pi/180.0)     # sample to detector distance in pixels
print("Sample to detector distance in pixels = " + str(det_R))
pylab.figure()
pylab.plot(pks, x*calib_tth_steps, 'b.')
pylab.plot(pks, lin_fit[0]*pks + lin_fit[1], 'r-')

outname = csv_path + csv_name[:-4] + "_calib.cal"
outfile = open(outname, "w")
outfile.write("direct_beam_x \t %i\n"  % db_pixel[0])
outfile.write("direct_beam_y \t %i\n" % db_pixel[1])
outfile.write("Sample_Detector_distance_pixels \t %15.6G\n" % det_R)
outfile.write("Sample_Detector_distance_mm \t %15.6G" % (det_R * pix_size / 1000.0))
outfile.close()

pylab.show()



# -*- coding: utf-8 -*-
"""
@author: walroth
"""

# Standard library imports

# Other imports
from pandas import Series

# Qt imports

# xdart imports
from xdart.utils import catch_h5py_file as catch

# This module imports
from .arch import EwaldArch


class ArchSeries():
    """Container for storing EwaldArch objects. Data is stored in an
    hdf5 file, rather than in memory. __getitem__ and __setitem__ have
    been overridden to write the information to a file, whose path is
    stored as an attribute.
    
    attributes:
        data_file: Path where hdf5 file is stored.
        file_lock: Thread safe lock which ensures only one thread
                   accesses data at a time.
        index: List of all arch id numbers.
    
    methods:
        append: Add a new arch at the end of the index.
        iloc: Retrieve an arch by its absolute location not its id.
        sort_index: Sort the index by arch id.
    """
    def __init__(self, data_file, file_lock, arches=[],
                 static=False, gi=False):
        """data_file: Path to hdf5 file for storing data.
        file_lock: Thread safe lock.
        arches: List of arches to initialize series with.
        """
        self.data_file = data_file
        self.file_lock = file_lock
        self.index = []
        self.static = static
        self.gi = gi
        if arches:
            for a in arches:
                self.__setitem__(a.idx, a)
        self._i = 0
        # invoke the lock to prevent conflicts
        with self.file_lock:
            # use catch to avoid oserrors which will resolve with time.
            with catch(self.data_file, 'a') as f:
                if 'arches' not in f:
                    f.create_group('arches')
                
    def __getitem__(self, idx):
        """Initializes a new EwaldArch object and loads data from file
        into it.
        """
        if idx in self.index:

            arch = EwaldArch(idx, static=self.static, gi=self.gi)
            # invoke the lock to prevent conflicts
            with self.file_lock:
                # use catch to avoid oserrors which will resolve with time.
                with catch(self.data_file, 'r') as f:
                    arch.load_from_h5(f['arches'])
            return arch
        else:
            raise KeyError(f"Arch not found with {idx} index")
    
    def iloc(self, idx):
        """Location based retrieval of arches instead of id based.
        Similar to .iloc in pandas Series but called by .iloc(i) not
        .iloc[i].
        """
        return self.__getitem__(self.index[idx])
    
    def __setitem__(self, idx, arch):
        """Sets the arch at location idx to be the new arch. If an arch
        is stored with the same idx, replaces the data with the new
        data.
        """
        # invoke the lock to prevent conflicts
        with self.file_lock:
            # use catch to avoid oserrors which will resolve with time.
            with catch(self.data_file, 'a') as f:
                if 'arches' not in f:
                    f.create_group('arches')
                if idx != arch.idx:
                    arch.idx = idx
                arch.save_to_h5(f['arches'])
                if arch.idx not in self.index:
                    self.index.append(arch.idx)
    
    def append(self, arch):
        """Adds a new arch to the end of the index, unless idx is
        already stored in which case the arch at that idx is replaced.
        """
        arches = ArchSeries(self.data_file, self.file_lock)
        arches.index = self.index[:]
        if isinstance(arch, Series):
            _arch = arch.iloc[0]
        else:
            _arch = arch
        arches.__setitem__(_arch.idx, _arch)
        return arches
    
    def sort_index(self, inplace=False):
        """Sorts the index by idx. If inplace is true, returns nothing.
        Else, returns a copy with the index sorted.
        """
        if inplace:
            self.index.sort()
        else:
            arches = ArchSeries(self.data_file, self.file_lock)
            arches.index = self.index[:]
            arches.index.sort()
            return arches
    
    def __next__(self):
        """Allows for iteration.
        """
        if self._i < len(self.index):
            arch = self.iloc(self._i)
            self._i += 1
            return arch
        else:
            raise StopIteration
    
    def __iter__(self):
        """Allows for iteration.
        """
        self._i = 0
        return self



import pickle
import sys
import warnings
from contextlib import contextmanager

from .smbase import SMBase, synced


class SMDict(SMBase):
    def __init__(self, addr=None, manager=None, mutex=None, size=0,
                 ratio=2) -> None:
        format_list = [
            '0'*128,
            int(size),
            int(size),
        ]
        super().__init__(addr, manager, mutex, format_list, size, ratio)

    @synced
    def cleanup(self) -> None:
        self._shm.close()

    @synced
    def move_to_end(self, key: str, last=True) -> None:
        warnings.warn(
            'The \'move_to_end\' method will be removed in future versions. '
            'Use pop and reassignment instead.',
            DeprecationWarning,
            stacklevel=2,
        )
        with self.get_dict() as db:
            db[key] = db.pop(key)

    @synced
    def clear(self) -> None:
        self._save_memory({})

    @synced
    def popitem(self, last=None):
        if last is not None:
            warnings.warn(
                'The \'last\' parameter will be removed in future versions. '
                'The \'popitem\' function now always returns last inserted.',
                DeprecationWarning,
                stacklevel=2,
            )
        with self.get_dict() as db:
            return db.popitem()

    @contextmanager
    def get_dict(self):
        db = self._read_memory()
        try:
            yield db
        finally:
            self._save_memory(db)

    @synced
    def __getitem__(self, key: str):
        return self._read_memory()[key]

    @synced
    def __setitem__(self, key: str, value) -> None:
        with self.get_dict() as db:
            db[key] = value

    @synced
    def __len__(self) -> int:
        return len(self._read_memory())

    @synced
    def __delitem__(self, key: str) -> None:
        with self.get_dict() as db:
            del db[key]

    @synced
    def __iter__(self):
        return iter(self._read_memory())

    @synced
    def __reversed__(self):
        return reversed(self._read_memory())

    @synced
    def __contains__(self, key: str) -> bool:
        return key in self._read_memory()

    @synced
    def __eq__(self, other) -> bool:
        return self._read_memory() == other

    @synced
    def __ne__(self, other) -> bool:
        return self._read_memory() != other

    if sys.version_info > (3, 8):
        @synced
        def __or__(self, other):
            return self._read_memory() | other

        @synced
        def __ror__(self, other):
            return other | self._read_memory()

        @synced
        def __ior__(self, other):
            with self.get_dict() as db:
                db |= other
                return db

    @synced
    def __str__(self) -> str:
        return str(self._read_memory())

    @synced
    def __repr__(self) -> str:
        return repr(self._read_memory())

    @synced
    def get(self, key: str, default=None):
        return self._read_memory().get(key, default)

    @synced
    def keys(self):  # type: ignore
        return self._read_memory().keys()

    @synced
    def values(self):  # type: ignore
        return self._read_memory().values()

    @synced
    def items(self):  # type: ignore
        return self._read_memory().items()

    @synced
    def pop(self, key: str, default=None):
        with self.get_dict() as db:
            if default is None:
                return db.pop(key)
            return db.pop(key, default)

    @synced
    def update(self, other=(), /, **kwds):
        with self.get_dict() as db:
            db.update(other, **kwds)

    @synced
    def setdefault(self, key: str, default=None):
        with self.get_dict() as db:
            return db.setdefault(key, default)

    def _save_memory(self, db) -> None:
        data = pickle.dumps(db, pickle.HIGHEST_PROTOCOL)
        self._shl[1] = len(data)
        if len(data) > self._shl[2]:
            self._recap(len(data) * 2)
        self._shm.buf[: len(data)] = data

    def _read_memory(self):
        if self._shl[1] == 0:
            return {}
        return pickle.loads(self._shm.buf[:self._shl[1]].tobytes())



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'h5viewerUI.ui'
#
# Created by: PyQt5 UI code generator 5.15.6
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(1440, 842)
        Form.setMinimumSize(QtCore.QSize(0, 0))
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setContentsMargins(8, 8, 8, 12)
        self.gridLayout.setObjectName("gridLayout")
        self.label_3 = QtWidgets.QLabel(Form)
        self.label_3.setMaximumSize(QtCore.QSize(16777215, 20))
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 1, 0, 1, 1)
        self.label_4 = QtWidgets.QLabel(Form)
        self.label_4.setMaximumSize(QtCore.QSize(70, 16777215))
        self.label_4.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.label_4.setObjectName("label_4")
        self.gridLayout.addWidget(self.label_4, 1, 1, 1, 1)
        self.show_all = QtWidgets.QPushButton(Form)
        self.show_all.setMaximumSize(QtCore.QSize(16777215, 25))
        self.show_all.setObjectName("show_all")
        self.gridLayout.addWidget(self.show_all, 3, 0, 1, 1)
        self.auto_last = QtWidgets.QPushButton(Form)
        self.auto_last.setMaximumSize(QtCore.QSize(16777215, 25))
        self.auto_last.setObjectName("auto_last")
        self.gridLayout.addWidget(self.auto_last, 3, 1, 1, 1)
        self.splitter = QtWidgets.QSplitter(Form)
        self.splitter.setMinimumSize(QtCore.QSize(0, 0))
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.splitter.setObjectName("splitter")
        self.listScans = QtWidgets.QListWidget(self.splitter)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.listScans.sizePolicy().hasHeightForWidth())
        self.listScans.setSizePolicy(sizePolicy)
        self.listScans.setMinimumSize(QtCore.QSize(30, 0))
        self.listScans.setResizeMode(QtWidgets.QListView.Adjust)
        self.listScans.setObjectName("listScans")
        self.listData = QtWidgets.QListWidget(self.splitter)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.listData.sizePolicy().hasHeightForWidth())
        self.listData.setSizePolicy(sizePolicy)
        self.listData.setMinimumSize(QtCore.QSize(45, 0))
        self.listData.setMaximumSize(QtCore.QSize(60, 16777215))
        self.listData.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.listData.setTabKeyNavigation(False)
        self.listData.setAlternatingRowColors(True)
        self.listData.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.listData.setObjectName("listData")
        self.gridLayout.addWidget(self.splitter, 2, 0, 1, 2)
        self.frame = QtWidgets.QFrame(Form)
        self.frame.setMaximumSize(QtCore.QSize(16777215, 20))
        self.frame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame.setObjectName("frame")
        self.gridLayout.addWidget(self.frame, 0, 0, 1, 2)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.label_3.setText(_translate("Form", "  Scans"))
        self.label_4.setText(_translate("Form", "Data"))
        self.show_all.setText(_translate("Form", "Show All"))
        self.auto_last.setText(_translate("Form", "Auto Last"))



import time
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler


class Watcher:
    DIRECTORY_TO_WATCH = "P:\\bl2-1"

    def __init__(self):
        self.observer = Observer()

    def run(self):
        event_handler = Handler()
        self.observer.schedule(event_handler, self.DIRECTORY_TO_WATCH, recursive=True)
        self.observer.start()
        try:
            while True:
                time.sleep(10)
        except:
            self.observer.stop()
            print("Error")

        self.observer.join()


class Handler(FileSystemEventHandler):

    @staticmethod
    def on_any_event(event):
        if event.is_directory:
            return None

        elif event.event_type == 'created':
            # Take any action here when a file is first created.
            print(f"Received created event - {event.src_path}")

        elif event.event_type == 'modified':
            # Taken any action here when a file is modified.
            print(f"Received modified event - {event.src_path}")


if __name__ == '__main__':
    w = Watcher()
    w.run()


import os, time
from fnmatch import filter

pdi_path = "P:\\bl2-1\\Dunn"
img_path = "Z:\\"

paths = [pdi_path, img_path]
filetypes = ['pdi', 'raw']

before = dict( [filetype, dict([(f, None) for f in filter(os.listdir (path), f'*{filetype}')])]
               for (path, filetype) in zip(paths, filetypes) )

while 1:
    time.sleep (0.1)
    after = dict ( [filetype, dict ([(f, None) for f in filter(os.listdir (path), f'*{filetype}')])]
               for (path, filetype) in zip(paths, filetypes) )
    
    for filetype in after.keys():
        added = [f for f in after[filetype] if not f in before[filetype]]
        removed = [f for f in before[filetype] if not f in after[filetype]]
        if added: print(f"{filetype} - Added: {''.join (added)}")
        if removed: print(f"{filetype} - Removed: {', '.join (removed)}")
    
    before = after


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'displayFrameUI.ui'
#
# Created by: PyQt5 UI code generator 5.15.6
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(1440, 814)
        self.layout = QtWidgets.QHBoxLayout(Form)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(0)
        self.layout.setObjectName("layout")
        self.splitter = QtWidgets.QSplitter(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.splitter.sizePolicy().hasHeightForWidth())
        self.splitter.setSizePolicy(sizePolicy)
        self.splitter.setOrientation(QtCore.Qt.Vertical)
        self.splitter.setHandleWidth(2)
        self.splitter.setObjectName("splitter")
        self.imageWindow = QtWidgets.QFrame(self.splitter)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.imageWindow.sizePolicy().hasHeightForWidth())
        self.imageWindow.setSizePolicy(sizePolicy)
        self.imageWindow.setMinimumSize(QtCore.QSize(0, 400))
        self.imageWindow.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.imageWindow.setFrameShadow(QtWidgets.QFrame.Raised)
        self.imageWindow.setLineWidth(3)
        self.imageWindow.setObjectName("imageWindow")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.imageWindow)
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_3.setSpacing(0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.frame_top = QtWidgets.QFrame(self.imageWindow)
        self.frame_top.setMinimumSize(QtCore.QSize(0, 35))
        self.frame_top.setMaximumSize(QtCore.QSize(16777215, 35))
        self.frame_top.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame_top.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_top.setObjectName("frame_top")
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout(self.frame_top)
        self.horizontalLayout_5.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_5.setSpacing(0)
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.frame_4 = QtWidgets.QFrame(self.frame_top)
        self.frame_4.setMinimumSize(QtCore.QSize(260, 0))
        self.frame_4.setMaximumSize(QtCore.QSize(260, 16777215))
        self.frame_4.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_4.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_4.setObjectName("frame_4")
        self.horizontalLayout_8 = QtWidgets.QHBoxLayout(self.frame_4)
        self.horizontalLayout_8.setContentsMargins(10, 0, 0, 0)
        self.horizontalLayout_8.setSpacing(0)
        self.horizontalLayout_8.setObjectName("horizontalLayout_8")
        self.horizontalLayout_7 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_7.setSpacing(0)
        self.horizontalLayout_7.setObjectName("horizontalLayout_7")
        self.normChannel = QtWidgets.QComboBox(self.frame_4)
        self.normChannel.setMinimumSize(QtCore.QSize(135, 0))
        self.normChannel.setMaximumSize(QtCore.QSize(140, 16777215))
        self.normChannel.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.normChannel.setObjectName("normChannel")
        self.normChannel.addItem("")
        self.normChannel.addItem("")
        self.normChannel.addItem("")
        self.normChannel.addItem("")
        self.normChannel.addItem("")
        self.normChannel.addItem("")
        self.horizontalLayout_7.addWidget(self.normChannel)
        spacerItem = QtWidgets.QSpacerItem(15, 20, QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_7.addItem(spacerItem)
        self.setBkg = QtWidgets.QPushButton(self.frame_4)
        self.setBkg.setMinimumSize(QtCore.QSize(90, 0))
        self.setBkg.setMaximumSize(QtCore.QSize(100, 16777215))
        self.setBkg.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.setBkg.setObjectName("setBkg")
        self.horizontalLayout_7.addWidget(self.setBkg)
        spacerItem1 = QtWidgets.QSpacerItem(10, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_7.addItem(spacerItem1)
        self.horizontalLayout_8.addLayout(self.horizontalLayout_7)
        self.horizontalLayout_5.addWidget(self.frame_4)
        self.frame_5 = QtWidgets.QFrame(self.frame_top)
        self.frame_5.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_5.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_5.setObjectName("frame_5")
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout(self.frame_5)
        self.horizontalLayout_6.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_6.setSpacing(0)
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.labelCurrent = QtWidgets.QLabel(self.frame_5)
        self.labelCurrent.setMaximumSize(QtCore.QSize(600, 16777215))
        font = QtGui.QFont()
        font.setPointSize(15)
        self.labelCurrent.setFont(font)
        self.labelCurrent.setAlignment(QtCore.Qt.AlignCenter)
        self.labelCurrent.setObjectName("labelCurrent")
        self.horizontalLayout_6.addWidget(self.labelCurrent)
        self.horizontalLayout_5.addWidget(self.frame_5)
        self.frame_6 = QtWidgets.QFrame(self.frame_top)
        self.frame_6.setMinimumSize(QtCore.QSize(260, 0))
        self.frame_6.setMaximumSize(QtCore.QSize(260, 16777215))
        self.frame_6.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_6.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_6.setObjectName("frame_6")
        self.horizontalLayout_10 = QtWidgets.QHBoxLayout(self.frame_6)
        self.horizontalLayout_10.setContentsMargins(0, 0, 10, 0)
        self.horizontalLayout_10.setSpacing(0)
        self.horizontalLayout_10.setObjectName("horizontalLayout_10")
        self.horizontalLayout_9 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_9.setSpacing(0)
        self.horizontalLayout_9.setObjectName("horizontalLayout_9")
        spacerItem2 = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_9.addItem(spacerItem2)
        self.scale = QtWidgets.QComboBox(self.frame_6)
        self.scale.setMinimumSize(QtCore.QSize(80, 0))
        self.scale.setMaximumSize(QtCore.QSize(80, 16777215))
        self.scale.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.scale.setObjectName("scale")
        self.scale.addItem("")
        self.scale.addItem("")
        self.scale.addItem("")
        self.scale.addItem("")
        self.horizontalLayout_9.addWidget(self.scale)
        spacerItem3 = QtWidgets.QSpacerItem(15, 20, QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_9.addItem(spacerItem3)
        self.cmap = QtWidgets.QComboBox(self.frame_6)
        self.cmap.setMinimumSize(QtCore.QSize(80, 0))
        self.cmap.setMaximumSize(QtCore.QSize(80, 16777215))
        self.cmap.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.cmap.setObjectName("cmap")
        self.cmap.addItem("")
        self.cmap.addItem("")
        self.cmap.addItem("")
        self.cmap.addItem("")
        self.cmap.addItem("")
        self.cmap.addItem("")
        self.cmap.addItem("")
        self.cmap.addItem("")
        self.cmap.addItem("")
        self.cmap.addItem("")
        self.cmap.addItem("")
        self.horizontalLayout_9.addWidget(self.cmap)
        self.horizontalLayout_10.addLayout(self.horizontalLayout_9)
        self.horizontalLayout_5.addWidget(self.frame_6)
        self.verticalLayout_3.addWidget(self.frame_top)
        self.twoDWindow = QtWidgets.QFrame(self.imageWindow)
        self.twoDWindow.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.twoDWindow.setFrameShadow(QtWidgets.QFrame.Raised)
        self.twoDWindow.setObjectName("twoDWindow")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout(self.twoDWindow)
        self.horizontalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_3.setSpacing(0)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.splitter_2 = QtWidgets.QSplitter(self.twoDWindow)
        self.splitter_2.setOrientation(QtCore.Qt.Horizontal)
        self.splitter_2.setHandleWidth(2)
        self.splitter_2.setObjectName("splitter_2")
        self.imageFrame = QtWidgets.QFrame(self.splitter_2)
        self.imageFrame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.imageFrame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.imageFrame.setObjectName("imageFrame")
        self.binnedFrame = QtWidgets.QFrame(self.splitter_2)
        self.binnedFrame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.binnedFrame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.binnedFrame.setObjectName("binnedFrame")
        self.horizontalLayout_3.addWidget(self.splitter_2)
        self.verticalLayout_3.addWidget(self.twoDWindow)
        self.imageToolbar = QtWidgets.QFrame(self.imageWindow)
        self.imageToolbar.setMinimumSize(QtCore.QSize(0, 40))
        self.imageToolbar.setMaximumSize(QtCore.QSize(16777215, 40))
        self.imageToolbar.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.imageToolbar.setFrameShadow(QtWidgets.QFrame.Raised)
        self.imageToolbar.setObjectName("imageToolbar")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.imageToolbar)
        self.horizontalLayout_2.setContentsMargins(12, 0, -1, 0)
        self.horizontalLayout_2.setSpacing(6)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.imageUnit = QtWidgets.QComboBox(self.imageToolbar)
        self.imageUnit.setMinimumSize(QtCore.QSize(90, 0))
        self.imageUnit.setMaximumSize(QtCore.QSize(130, 16777215))
        self.imageUnit.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.imageUnit.setObjectName("imageUnit")
        self.imageUnit.addItem("")
        self.imageUnit.addItem("")
        self.imageUnit.addItem("")
        self.horizontalLayout_2.addWidget(self.imageUnit)
        spacerItem4 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem4)
        self.update2D = QtWidgets.QCheckBox(self.imageToolbar)
        self.update2D.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.update2D.setChecked(True)
        self.update2D.setObjectName("update2D")
        self.horizontalLayout_2.addWidget(self.update2D)
        spacerItem5 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem5)
        self.shareAxis = QtWidgets.QCheckBox(self.imageToolbar)
        self.shareAxis.setMaximumSize(QtCore.QSize(90, 16777215))
        self.shareAxis.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.shareAxis.setObjectName("shareAxis")
        self.horizontalLayout_2.addWidget(self.shareAxis)
        spacerItem6 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem6)
        self.save_2D = QtWidgets.QPushButton(self.imageToolbar)
        self.save_2D.setObjectName("save_2D")
        self.horizontalLayout_2.addWidget(self.save_2D)
        self.verticalLayout_3.addWidget(self.imageToolbar)
        self.plotWindow = QtWidgets.QFrame(self.splitter)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.plotWindow.sizePolicy().hasHeightForWidth())
        self.plotWindow.setSizePolicy(sizePolicy)
        self.plotWindow.setMinimumSize(QtCore.QSize(0, 400))
        self.plotWindow.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.plotWindow.setFrameShadow(QtWidgets.QFrame.Raised)
        self.plotWindow.setLineWidth(3)
        self.plotWindow.setObjectName("plotWindow")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.plotWindow)
        self.verticalLayout_4.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_4.setSpacing(0)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.oneDWindow = QtWidgets.QFrame(self.plotWindow)
        self.oneDWindow.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.oneDWindow.setFrameShadow(QtWidgets.QFrame.Raised)
        self.oneDWindow.setObjectName("oneDWindow")
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout(self.oneDWindow)
        self.horizontalLayout_4.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_4.setSpacing(0)
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.splitter_3 = QtWidgets.QSplitter(self.oneDWindow)
        self.splitter_3.setOrientation(QtCore.Qt.Horizontal)
        self.splitter_3.setHandleWidth(2)
        self.splitter_3.setObjectName("splitter_3")
        self.plotFrame = QtWidgets.QFrame(self.splitter_3)
        self.plotFrame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.plotFrame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.plotFrame.setObjectName("plotFrame")
        self.horizontalLayout_4.addWidget(self.splitter_3)
        self.verticalLayout_4.addWidget(self.oneDWindow)
        self.plotToolBar = QtWidgets.QFrame(self.plotWindow)
        self.plotToolBar.setMinimumSize(QtCore.QSize(0, 40))
        self.plotToolBar.setMaximumSize(QtCore.QSize(16777215, 40))
        self.plotToolBar.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.plotToolBar.setFrameShadow(QtWidgets.QFrame.Raised)
        self.plotToolBar.setObjectName("plotToolBar")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.plotToolBar)
        self.horizontalLayout.setContentsMargins(12, 0, -1, 0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.plotUnit = QtWidgets.QComboBox(self.plotToolBar)
        self.plotUnit.setMinimumSize(QtCore.QSize(70, 0))
        self.plotUnit.setMaximumSize(QtCore.QSize(100, 16777215))
        self.plotUnit.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.plotUnit.setObjectName("plotUnit")
        self.plotUnit.addItem("")
        self.plotUnit.addItem("")
        self.plotUnit.addItem("")
        self.horizontalLayout.addWidget(self.plotUnit)
        self.slice = QtWidgets.QCheckBox(self.plotToolBar)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.slice.sizePolicy().hasHeightForWidth())
        self.slice.setSizePolicy(sizePolicy)
        self.slice.setMinimumSize(QtCore.QSize(80, 0))
        self.slice.setMaximumSize(QtCore.QSize(80, 16777215))
        self.slice.setObjectName("slice")
        self.horizontalLayout.addWidget(self.slice)
        self.slice_center = QtWidgets.QDoubleSpinBox(self.plotToolBar)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.slice_center.sizePolicy().hasHeightForWidth())
        self.slice_center.setSizePolicy(sizePolicy)
        self.slice_center.setMinimumSize(QtCore.QSize(70, 0))
        self.slice_center.setMaximumSize(QtCore.QSize(70, 16777215))
        self.slice_center.setToolTip("")
        self.slice_center.setToolTipDuration(2)
        self.slice_center.setWhatsThis("")
        self.slice_center.setAccessibleDescription("")
        self.slice_center.setMinimum(-180.0)
        self.slice_center.setMaximum(180.0)
        self.slice_center.setSingleStep(0.5)
        self.slice_center.setProperty("value", 0.0)
        self.slice_center.setObjectName("slice_center")
        self.horizontalLayout.addWidget(self.slice_center)
        self.slice_width = QtWidgets.QDoubleSpinBox(self.plotToolBar)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.slice_width.sizePolicy().hasHeightForWidth())
        self.slice_width.setSizePolicy(sizePolicy)
        self.slice_width.setMinimumSize(QtCore.QSize(70, 0))
        self.slice_width.setMaximumSize(QtCore.QSize(70, 16777215))
        self.slice_width.setToolTip("")
        self.slice_width.setMinimum(0.0)
        self.slice_width.setMaximum(270.0)
        self.slice_width.setSingleStep(0.5)
        self.slice_width.setProperty("value", 5.0)
        self.slice_width.setObjectName("slice_width")
        self.horizontalLayout.addWidget(self.slice_width)
        spacerItem7 = QtWidgets.QSpacerItem(80, 20, QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem7)
        self.plotMethod = QtWidgets.QComboBox(self.plotToolBar)
        self.plotMethod.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.plotMethod.setObjectName("plotMethod")
        self.plotMethod.addItem("")
        self.plotMethod.addItem("")
        self.plotMethod.addItem("")
        self.plotMethod.addItem("")
        self.plotMethod.addItem("")
        self.horizontalLayout.addWidget(self.plotMethod)
        self.yOffsetLabel = QtWidgets.QLabel(self.plotToolBar)
        self.yOffsetLabel.setMaximumSize(QtCore.QSize(140, 16777215))
        self.yOffsetLabel.setObjectName("yOffsetLabel")
        self.horizontalLayout.addWidget(self.yOffsetLabel)
        self.yOffset = QtWidgets.QDoubleSpinBox(self.plotToolBar)
        self.yOffset.setEnabled(False)
        self.yOffset.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.yOffset.setDecimals(1)
        self.yOffset.setMinimum(-100.0)
        self.yOffset.setSingleStep(5.0)
        self.yOffset.setProperty("value", 5.0)
        self.yOffset.setObjectName("yOffset")
        self.horizontalLayout.addWidget(self.yOffset)
        self.wf_options = QtWidgets.QPushButton(self.plotToolBar)
        self.wf_options.setEnabled(False)
        self.wf_options.setToolTip("")
        self.wf_options.setWhatsThis("")
        self.wf_options.setObjectName("wf_options")
        self.horizontalLayout.addWidget(self.wf_options)
        spacerItem8 = QtWidgets.QSpacerItem(30, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem8)
        self.showLegend = QtWidgets.QCheckBox(self.plotToolBar)
        self.showLegend.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.showLegend.setChecked(True)
        self.showLegend.setObjectName("showLegend")
        self.horizontalLayout.addWidget(self.showLegend)
        self.clear_1D = QtWidgets.QPushButton(self.plotToolBar)
        self.clear_1D.setObjectName("clear_1D")
        self.horizontalLayout.addWidget(self.clear_1D)
        self.save_1D = QtWidgets.QPushButton(self.plotToolBar)
        self.save_1D.setObjectName("save_1D")
        self.horizontalLayout.addWidget(self.save_1D)
        self.verticalLayout_4.addWidget(self.plotToolBar)
        self.layout.addWidget(self.splitter)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)
        Form.setTabOrder(self.normChannel, self.setBkg)
        Form.setTabOrder(self.setBkg, self.scale)
        Form.setTabOrder(self.scale, self.cmap)
        Form.setTabOrder(self.cmap, self.imageUnit)
        Form.setTabOrder(self.imageUnit, self.save_2D)
        Form.setTabOrder(self.save_2D, self.plotUnit)
        Form.setTabOrder(self.plotUnit, self.slice_center)
        Form.setTabOrder(self.slice_center, self.slice_width)
        Form.setTabOrder(self.slice_width, self.plotMethod)
        Form.setTabOrder(self.plotMethod, self.yOffset)
        Form.setTabOrder(self.yOffset, self.save_1D)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.normChannel.setItemText(0, _translate("Form", "Norm Channel"))
        self.normChannel.setItemText(1, _translate("Form", "Monitor"))
        self.normChannel.setItemText(2, _translate("Form", "sec"))
        self.normChannel.setItemText(3, _translate("Form", "bstop"))
        self.normChannel.setItemText(4, _translate("Form", "I0"))
        self.normChannel.setItemText(5, _translate("Form", "I1"))
        self.setBkg.setText(_translate("Form", "Set Bkg"))
        self.labelCurrent.setText(_translate("Form", "Current"))
        self.scale.setItemText(0, _translate("Form", "Linear"))
        self.scale.setItemText(1, _translate("Form", "Log"))
        self.scale.setItemText(2, _translate("Form", "Log-Log"))
        self.scale.setItemText(3, _translate("Form", "Sqrt"))
        self.cmap.setItemText(0, _translate("Form", "Default"))
        self.cmap.setItemText(1, _translate("Form", "viridis"))
        self.cmap.setItemText(2, _translate("Form", "grey"))
        self.cmap.setItemText(3, _translate("Form", "plasma"))
        self.cmap.setItemText(4, _translate("Form", "inferno"))
        self.cmap.setItemText(5, _translate("Form", "magma"))
        self.cmap.setItemText(6, _translate("Form", "thermal"))
        self.cmap.setItemText(7, _translate("Form", "flame"))
        self.cmap.setItemText(8, _translate("Form", "yellowy"))
        self.cmap.setItemText(9, _translate("Form", "bipolar"))
        self.cmap.setItemText(10, _translate("Form", "greyclip"))
        self.imageUnit.setItemText(0, _translate("Form", "Q-Chi"))
        self.imageUnit.setItemText(1, _translate("Form", "2Th-Chi"))
        self.imageUnit.setItemText(2, _translate("Form", "Qz-Qxy"))
        self.update2D.setText(_translate("Form", "Update 2D"))
        self.shareAxis.setText(_translate("Form", "Share Axis"))
        self.save_2D.setText(_translate("Form", "Save"))
        self.plotUnit.setItemText(0, _translate("Form", "Q (A-1)"))
        self.plotUnit.setItemText(1, _translate("Form", "2 u\\u03B8"))
        self.plotUnit.setItemText(2, _translate("Form", "Chi"))
        self.slice.setText(_translate("Form", "X Range"))
        self.plotMethod.setItemText(0, _translate("Form", "Single"))
        self.plotMethod.setItemText(1, _translate("Form", "Overlay"))
        self.plotMethod.setItemText(2, _translate("Form", "Average"))
        self.plotMethod.setItemText(3, _translate("Form", "Sum"))
        self.plotMethod.setItemText(4, _translate("Form", "Waterfall"))
        self.yOffsetLabel.setToolTip(_translate("Form", "y Offset for Overlay Mode"))
        self.yOffsetLabel.setText(_translate("Form", "Offset"))
        self.wf_options.setText(_translate("Form", "Options"))
        self.showLegend.setText(_translate("Form", "Legend"))
        self.clear_1D.setText(_translate("Form", "Clear"))
        self.save_1D.setText(_translate("Form", "Save"))



from .typedefs import int_t, dtype_to_key, key_to_dtype
from ._smutils import none_lesser, none_greater, multiply_none
from .smbase import SMBase


class SMVector(SMBase):
    def __init__(self, addr=None, manager=None, mutex=None, size=0, dtype=int_t,
                 ratio=2.0):
        format_list = [
            '0'*128,
            int(size),
            int(size),
            '0' * 16,
            float(ratio)
        ]
        super().__init__(addr, manager, mutex, format_list, size*dtype.size,
                         int(ratio*dtype.size))
        with self.mutex:
            if self._shl[3] == '0' * 16:
                self._shl[3] = dtype_to_key[dtype]
            self._dtype = key_to_dtype[self._shl[3]]
            self._shl[4] = ratio

    def size(self):
        with self.mutex:
            out = self._shl[1]
        return out//self._dtype.size

    def capacity(self):
        with self.mutex:
            out = self._shl[2]
        return out//self._dtype.size

    def dtype(self):
        with self.mutex:
            out = key_to_dtype[self._shl[3]]
        return out

    def push_back(self, val):
        with self.mutex:
            self.check_memory()
            if self._shl[1] == self._shl[2]:
                self._recap(self._shm.size*self._shl[4])
            idx = self._shl[1]
            self._shm.buf[idx:idx + self._dtype.size] = self._dtype.pack(val)
            self._shl[1] += self._dtype.size

    def _check_key(self, key):
        with self.mutex:
            if isinstance(key, slice):
                if none_greater(key.start, self._shl[1]) or none_lesser(key.start, (-self._shl[1])) or \
                        none_greater(key.stop, self._shl[1]) or none_lesser(key.stop, (-self._shl[1])):
                    raise IndexError("Index out of range")
                start = multiply_none(key.start, self._dtype.size)
                stop = multiply_none(key.stop, self._dtype.size)
                if start is not None:
                    if start < 0:
                        start += self._shl[1]
                else:
                    start = 0
                if stop is not None:
                    if stop < 0:
                        stop += self._shl[1]
                else:
                    stop = self._shl[1]
                return slice(
                    start,
                    stop,
                    multiply_none(key.step, self._dtype.size)
                )
            elif isinstance(key, int):
                size = self._shl[1]//self._dtype.size
                if key >= size or key < (-size):
                    raise IndexError("Index out of range")
                return key * self._dtype.size
            else:
                raise TypeError(f"index must be slice or int, not {type(key)}")

    def __getitem__(self, key):
        with self.mutex:
            self.check_memory()
            adj_key = self._check_key(key)
            if isinstance(adj_key, slice):
                out_bytes = self._shm.buf[adj_key]
                out = [self._dtype.unpack(out_bytes[i: i + self._dtype.size])
                       for i in range(0, len(out_bytes), self._dtype.size)]
                out_bytes.release()
            else:
                if adj_key < 0:
                    adj_key += self._shl[1]
                out = self._dtype.unpack(self._shm.buf[adj_key:adj_key + self._dtype.size])

        return out

    def __setitem__(self, key, value):
        with self.mutex:
            self.check_memory()
            adj_key = self._check_key(key)
            if isinstance(adj_key, slice):
                self._shm.buf[adj_key] = b''.join([self._dtype.pack(x) for x in value])
            else:
                if adj_key < 0:
                    adj_key += self._shl[1]
                self._shm.buf[adj_key:adj_key + self._dtype.size] = self._dtype.pack(value)



import socket
import time
import types
import sys
import yaml
import copy
import logging
from logging import info, warning, critical, debug
import asyncio



class BLError(Exception):
    """This is the base class for a broader range of errors coming from the MDrive motors"""
    def __init__(self, expression=None, message=None, fatalError=True, canRestart=True):
        self.expression = expression
        self.message = message
        full_message = "{}: {}".format(expression, message)
        warning(full_message)
        print(full_message)

class BLCommandError(BLError):
    pass


class SPECCommandError(BLCommandError):
    pass





class BLConfigFileError(BLError):
    pass


class SPECError(BLError):
    pass



class BLCommandError(BLError):
    pass

class SPECCommunicationError(BLCommandError):
    pass


class SISControlError(BLError):
    pass



from multiprocessing.managers import SharedMemoryManager
from multiprocessing import shared_memory, RLock
import traceback
from functools import wraps

from ._smutils import DummyManager


def synced(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        with self.mutex:
            self.check_memory()
            out = func(self, *args, **kwargs)
        return out
    return wrapper


def locked(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        with self.mutex:
            out = func(self, *args, **kwargs)
        return out
    return wrapper


class SMBase:
    def __init__(self, addr=None, manager=None, mutex=None, format_list=None,
                 size=0, ratio=2):
        if mutex is None:
            self.mutex = RLock()
        else:
            self.mutex = mutex
        if manager is None:
            self._manager = DummyManager()
        elif isinstance(manager, SharedMemoryManager) or isinstance(manager, DummyManager):
            self._manager = manager
        else:
            self._manager = SharedMemoryManager(manager)
        capacity = int(size*ratio)
        if capacity == 0:
            capacity = ratio
        with self.mutex:
            if addr is None:
                if format_list is None:
                    raise ValueError("Either addr or format_list must be provided")
                self._shl = self._manager.ShareableList(format_list)
                self._shl_addr = self._shl.shm.name
                self._shm = self._manager.SharedMemory(capacity)
                self._shm_addr = self._shm.name
                self._shl[0] = self._shm_addr
                self._shl[1] = int(size)
                self._shl[2] = int(capacity)
            else:
                self._shl_addr = addr
                self._shl = shared_memory.ShareableList(name=addr)
                self._shm_addr = self._shl[0]
                self._shm = shared_memory.SharedMemory(name=self._shm_addr)

    @locked
    def size(self):
        return self._shl[1]

    @locked
    def capacity(self):
        return self._shl[2]

    @locked
    def name(self):
        return self._shl.shm.name

    def check_memory(self) -> bool:
        updated = False
        with self.mutex:
            if self._shm_addr != self._shl[0]:
                updated = True
                try:
                    self._shm.close()
                except:
                    traceback.print_exc()
                self._shm = shared_memory.SharedMemory(name=self._shl[0])
                self._shm_addr = self._shl[0]
        return updated

    def _recap(self, cap: int):
        if cap > self._shm.size:
            new_shm = self._manager.SharedMemory(size=int(cap))
            new_shm.buf[:self._shm.size] = self._shm.buf[:]
            try:
                self._shm.close()
                self._shm.unlink()
            except:
                traceback.print_exc()
            self._shm = new_shm
            self._shm_addr = new_shm.name
            self._shl[0] = new_shm.name
            self._shl[2] = int(cap)

    def __del__(self):
        """
        Note: this does not unlink data. That is expected to be handled
        by the manager.
        """
        self._shm.close()
        self._shl.shm.close()



# -*- coding: utf-8 -*-
"""
@author: walroth
"""

# Standard library imports

# Other imports
import pandas as pd

# Qt imports

# This module imports
from xdart.utils import soft_list_eval


def get_spec_header(file_path):
    """Gets the header information of a spec data file.
    
    args:
        file_path: str, path to spec data file.
    
    returns:
        header: dict, header information including metadata and motor
            and detector names.
    """
    cont = True
    header = {
        'meta': {},
        'motors': {},
        'motors_r': {},
        'detectors': {},
        'detectors_r': {},
    }
    with open(file_path, 'r') as file:
        for lin in file:
            line = lin.split()
            if line:
                key = line[0]
                if '#S' in key:
                    break
                code = line[0][1:]
                if code == 'F': 
                    header['meta']['File'] = line[1:]

                elif code == 'E':
                    header['meta']['Epoch'] = eval(line[1])

                elif code == 'D':
                    header['meta']['Date'] = ' '.join(line[1:])

                elif code == 'C':
                    header['meta']['Comment'] = ' '.join(line[1:])
                    for i, val in enumerate(line):
                        if val == 'User' and line[i + 1] == '=':
                            header['meta']['User'] = line[i+2]
                
                elif 'O' in code:
                    num = int(code[1:])
                    header['motors'][num] = line[1:]
                
                elif 'o' in code:
                    num = int(code[1:])
                    header['motors_r'][num] = line[1:]
                
                elif 'J' in code:
                    num = int(code[1:])
                    header['detectors'][num] = line[1:]
                
                elif 'j' in code:
                    num = int(code[1:])
                    header['detectors_r'][num] = line[1:]
    return header


def get_spec_scan(file_path, scan_number, header):
    """Reads a spec data file and returns a specified scan as a
    pandas DataFrame.
    
    args:
        file_path: str, path to spec data file.
        scan_number: int, scan to look for.
        header: dict, header information returned from get_header.
    
    returns:
        df: pandas DataFrame, the data from the scan.
        meta: dict, metadata associated with scan.
    """
    cont = True
    df = None
    
    with open(file_path, 'r') as file:
        for lin in file:
            line = lin.split()
            if not line:
                if cont:
                    continue
                else:
                    break
            else:
                if line[0] == '#S' and str(scan_number) == line[1]:
                    cont = False
                    df, meta = _parse_scan(line, header)
                elif not cont:
                    df, meta = _parse_scan(line, header, df, meta)
    if df is None:
        raise KeyError("Scan not found")
    return df, meta


def _parse_scan(line, header, df=None, meta=None):
    """Helper function to parse information from spec file into scan
    data.
    
    args:
        header: dict, header information returned from get_header.
        df: pandas DataFrame, scan data read to this point in scan.
        meta: dict, metadata associated with the scan.
    
    returns:
        df: pandas DataFrame, the data from the scan.
        meta: dict, metadata associated with scan.
    """
    if meta is None:
        meta = {}
        
    flag = line[0]
    if '#' in flag:
        if 'S' in flag:
            meta.update({'Goniometer': {}, 'Motors': {}})
            meta['Command'] = ' '.join(line[2:])
            meta['Type'] = line[2]

        elif 'D' in flag:
            meta['Date'] = ' '.join(line[1:])

        elif 'T' in flag or 'M' in flag:
            meta['Counter'] = {
                'Amount': eval(line[1]), 'Type': line[2]
            }

        elif 'G' in flag:
            key = int(flag[2:])
            meta['Goniometer'][key] = soft_list_eval(line[1:])

        elif 'Q' in flag:
            meta['HKL'] = soft_list_eval(line[1:])

        elif 'P' in flag:
            motor_num = int(flag[2:])
            names = header['motors'][motor_num]
            positions = [eval(x) for x in line[1:]]
            meta['Motors'].update(
                {name: position for name, position in 
                zip(names, positions)}
            )
        # TODO: decide what to do with N;

        elif 'L' in flag:
            df = pd.DataFrame(columns=line[1:])

    else:
        vals = soft_list_eval(line)
        try:
            idx = df.index[-1] + 1
            df.loc[idx] = vals
        except IndexError:
            df.loc[0] = vals
    return df, meta



# -*- coding: utf-8 -*-
"""
@author: walroth
"""
# Standard library imorts
import json
import os
import traceback

# Other imports
from xdart.utils import catch_h5py_file as catch

# Qt imports
from pyqtgraph import Qt
from pyqtgraph.Qt import QtWidgets

QTreeWidget = QtWidgets.QTreeWidget
QTreeWidgetItem = QtWidgets.QTreeWidgetItem
QWidget = QtWidgets.QWidget
QFileDialog = QtWidgets.QFileDialog

# This module imports
from xdart.utils import catch_h5py_file
from .h5viewerUI import Ui_Form
from .sphere_threads import fileHandlerThread
from ...widgets import defaultWidget
from ...gui_utils import XdartDecoder, XdartEncoder
from pathlib import Path

class H5Viewer(QWidget):
    """Widget for displaying the contents of an EwaldSphere object and
    a basic file explorer. Also holds menus for more general tasks like
    setting defaults.
    
    attributes:
        (QAction attributes not shown, associated menus are)
        exportMenu: Sub-menu for exporting images and 1d data
        file_lock: Condition, lock governing file access
        fileMenu: Menu for saving files and exporting data
        fname: Current data file name
        layout: ui layout TODO: this can stay with ui
        paramMenu: Menu for saving and loading defaults
        toolbar: QToolBar, holds the menus
        ui: Ui_Form from qtdesigner

    methods:
        set_data: Sets the data in the dataList
        set_open_enabled: Sets the ability to open scans to enabled or
            disables
        update: Updates files in scansList
        TODO: Rename the methods and attributes based on what they
            actually do
    """
    sigNewFile = Qt.QtCore.Signal(str)
    sigUpdate = Qt.QtCore.Signal()
    sigThreadFinished = Qt.QtCore.Signal()

    def __init__(self, file_lock, local_path, dirname, sphere, arch, parent=None):
        super().__init__(parent)
        self.local_path = local_path
        self.file_lock = file_lock
        self.dirname = dirname
        if self.dirname == '':
            self.dirname = str(Path.home())
        self.sphere = sphere
        self.arch = arch
        self.ui = Ui_Form()
        self.ui.setupUi(self)

        self.layout = self.ui.layout
        self.defaultWidget = defaultWidget()
        self.defaultWidget.sigSetUserDefaults.connect(self.set_user_defaults)

        # Toolbar setup
        self.toolbar = QtWidgets.QToolBar('Tools')

        # File menu setup (part of toolbar)
        self.actionOpenFolder = QtWidgets.QAction()
        self.actionOpenFolder.setText('Open Folder')

        self.actionSetDefaults = QtWidgets.QAction()
        self.actionSetDefaults.setText('Advanced...')

        self.actionSaveDataAs = QtWidgets.QAction()
        self.actionSaveDataAs.setText('Save As')

        self.actionNewFile = QtWidgets.QAction()
        self.actionNewFile.setText('New')

        self.exportMenu = QtWidgets.QMenu()
        self.exportMenu.setTitle('Export')

        self.actionSaveImage = QtWidgets.QAction()
        self.actionSaveImage.setText('Current Image')
        self.exportMenu.addAction(self.actionSaveImage)

        self.actionSaveArray = QtWidgets.QAction()
        self.actionSaveArray.setText('Current 1D Array')
        self.exportMenu.addAction(self.actionSaveArray)
        
        self.paramMenu = QtWidgets.QMenu()
        self.paramMenu.setTitle('Config')
        
        self.actionSaveParams = QtWidgets.QAction()
        self.actionSaveParams.setText('Save')
        self.actionSaveParams.triggered.connect(self.defaultWidget.save_defaults)
        self.paramMenu.addAction(self.actionSaveParams)
        
        self.actionLoadParams = QtWidgets.QAction()
        self.actionLoadParams.setText('Load')
        self.actionLoadParams.triggered.connect(self.defaultWidget.load_defaults)
        self.paramMenu.addAction(self.actionLoadParams)
        
        self.paramMenu.addAction(self.actionSetDefaults)

        self.fileMenu = QtWidgets.QMenu()
        self.fileMenu.addAction(self.actionOpenFolder)
        self.fileMenu.addAction(self.actionNewFile)
        self.fileMenu.addAction(self.actionSaveDataAs)
        self.fileMenu.addMenu(self.exportMenu)

        self.fileButton = QtWidgets.QToolButton()
        self.fileButton.setText('File')
        self.fileButton.setPopupMode(QtWidgets.QToolButton.InstantPopup)
        self.fileButton.setMenu(self.fileMenu)
        
        self.paramButton = QtWidgets.QToolButton()
        self.paramButton.setText('Config')
        self.paramButton.setPopupMode(QtWidgets.QToolButton.InstantPopup)
        self.paramButton.setMenu(self.paramMenu)
        # End file menu setup

        self.toolbar.addWidget(self.fileButton)
        self.toolbar.addWidget(self.paramButton)
        # End toolbar setup

        self.layout.addWidget(self.toolbar, 0, 0, 1, 2)
        self.actionSetDefaults.triggered.connect(self.defaultWidget.show)
        
        self.ui.listScans.itemDoubleClicked.connect(self.scans_clicked)
        self.ui.listData.currentItemChanged.connect(self.data_clicked)
        self.actionOpenFolder.triggered.connect(self.open_folder)
        self.actionSaveDataAs.triggered.connect(self.save_data_as)
        self.actionNewFile.triggered.connect(self.new_file)
        
        self.file_thread = fileHandlerThread(self.sphere, self.arch,
                                             self.file_lock)
        self.file_thread.sigTaskDone.connect(self.thread_finished)
        self.file_thread.sigNewFile.connect(self.sigNewFile.emit)
        self.file_thread.sigUpdate.connect(self.sigUpdate.emit)
        self.file_thread.start(Qt.QtCore.QThread.LowPriority)
        
        self.update()
        self.show()

    def load_starting_defaults(self):
        default_path = os.path.join(self.local_path, "last_defaults.json")
        if os.path.exists(default_path):
            self.defaultWidget.load_defaults(fname=default_path)
        else:
            self.defaultWidget.save_defaults(fname=default_path)

    def set_user_defaults(self):
        default_path = os.path.join(self.local_path, "last_defaults.json")
        self.defaultWidget.save_defaults(fname=default_path)


    def update(self):
        """Calls both update_scans and update_data.
        """
        self.update_scans()
        self.update_data()
        
    def update_scans(self):
        """Takes in directory path and adds files in path to listScans
        """
        self.ui.listScans.clear()
        self.ui.listScans.addItem('..')
        if self.dirname == '':
            self.dirname = str(Path.home())
        for name in os.listdir(self.dirname):
            abspath = os.path.join(self.dirname, name)
            if os.path.isdir(abspath):
                self.ui.listScans.addItem(name + '/')
            elif name.split('.')[-1] in ('h5', 'hdf5'):
                self.ui.listScans.addItem(name)
    
    def update_data(self):
        """Updates list with all arch ids.
        """
        previous_loc = self.ui.listData.currentRow()
        if self.sphere.name != "null_main":
            with self.sphere.sphere_lock:
                _idxs = list(self.sphere.arches.index)
            self.ui.listData.clear()
            self.ui.listData.addItem('Overall')
            for idx in _idxs:
                self.ui.listData.addItem(str(idx))
        if previous_loc > self.ui.listData.count() - 1:
            previous_loc = self.ui.listData.count() - 1
        self.ui.listData.setCurrentRow(previous_loc)
    
    def thread_finished(self, task):
        if task != "load_arch":
            self.update()
        self.sigThreadFinished.emit()
    
    def scans_clicked(self, q):
        """Handles items being double clicked in listScans. Either
        navigates to new folder or loads in sphere data.
        
        q: QListItem, item selected in h5viewer.
        """
        if q.data(0) == '..':
            if self.dirname[-1] in ['/', '\\']:
                up = os.path.dirname(self.dirname[:-1])
            else:
                up = os.path.dirname(self.dirname)
            
            if os.path.isdir(up) and os.path.splitdrive(up)[1] != '':
                self.dirname = up
                self.update_scans()
        elif '/' in q.data(0):
            dirname = os.path.join(self.dirname, q.data(0))
            if os.path.isdir(dirname):
                self.dirname = dirname
                self.update_scans()
        elif q.data(0) != 'No scans':
            self.set_file(os.path.join(self.dirname, q.data(0)))
    
    def set_file(self, fname):
        """Changes the data file.
        
        args:
            fname: str, absolute path for data file
        """
        if fname != '':
            try:
                with self.file_lock:
                    with catch_h5py_file(fname, 'a') as _:
                        pass
            
                self.ui.listData.clear()
                self.ui.listData.addItem('Loading...')
                self.set_open_enabled(False)
                self.file_thread.fname = fname
                self.file_thread.queue.put("set_datafile")
            except:
                traceback.print_exc()
                return
    
    def data_clicked(self, current, previous):
        """Connected to currentItemChanged signal of listData
        
        current: QListItem, item selected
        previous: QListItem, previous selection
        """
        if current is not None and previous is not None:
            nochange = (current.data(0) == previous.data(0))
        else:
            nochange = False
        if (current is not None and 
                current.data(0) not in ('No data', "Loading...") and not
                nochange):
            self.arch.reset()
            if current.data(0) != 'Overall' and 'scan' not in current.data(0):
                try:
                    idx = int(current.data(0))
                except ValueError:
                    return
                self.arch.idx = idx
                self.file_thread.queue.put("load_arch")
            else:
                self.sigUpdate.emit()
    
    def open_folder(self):
        """Changes the directory being displayed in the file explorer.
        """
        dirname = QFileDialog().getExistingDirectory()
        if dirname != "":
            self.dirname = dirname
            self.update_scans()
    
    def set_open_enabled(self, enable):
        """Sets the save and open actions to enable
        
        args:
            enable: bool, if True actions are enabled
        """
        self.actionSaveDataAs.setEnabled(enable)
        self.paramMenu.setEnabled(enable)
        self.actionOpenFolder.setEnabled(enable)
        self.actionNewFile.setEnabled(enable)
        self.ui.listScans.setEnabled(enable)
    
    def save_data_as(self):
        """Saves all data to hdf5 file. Also sets fname to be the
        selected file.
        """
        fname, _ = QFileDialog.getSaveFileName()
        with self.file_thread.lock:
            self.file_thread.new_fname = fname
            self.file_thread.queue.put("save_data_as")
        self.set_file(fname)
    
    def new_file(self):
        """Calls file dialog and sets the file name.
        """
        fname, _ = QFileDialog.getSaveFileName()
        self.set_file(fname)



# -*- coding: utf-8 -*-
"""
@author: walroth
"""
# Standard library imports
import os
import time
import traceback
import gc

# This module imports
from .ui.h5viewerUI import Ui_Form
from xdart.modules.ewald import EwaldArch
from .sphere_threads import fileHandlerThread
from ...widgets import defaultWidget
from xdart import utils
from xdart.utils.containers import int_2d_data_static
from xdart.utils import catch_h5py_file as catch
from xdart.utils.containers import create_ai_from_dict

# Qt imports
from pyqtgraph import Qt
from pyqtgraph.Qt import QtWidgets, QtCore

# from icecream import ic; ic.configureOutput(prefix='', includeContext=True)

QTreeWidget = QtWidgets.QTreeWidget
QTreeWidgetItem = QtWidgets.QTreeWidgetItem
QWidget = QtWidgets.QWidget
QFileDialog = QtWidgets.QFileDialog
QItemSelectionModel = QtCore.QItemSelectionModel


class H5Viewer(QWidget):
    """Widget for displaying the contents of an EwaldSphere object and
    a basic file explorer. Also holds menus for more general tasks like
    setting defaults.
    
    attributes:
        (QAction attributes not shown, associated menus are)
        exportMenu: Sub-menu for exporting images and 1d data
        file_lock: Condition, lock governing file access
        fileMenu: Menu for saving files and exporting data
        fname: Current data file name
        layout: ui layout TODO: this can stay with ui
        paramMenu: Menu for saving and loading defaults
        toolbar: QToolBar, holds the menus
        ui: Ui_Form from qtdesigner

    methods:
        set_data: Sets the data in the dataList
        set_open_enabled: Sets the ability to open scans to enabled or
            disables
        update: Updates files in scansList
        TODO: Rename the methods and attributes based on what they
            actually do
    """
    sigNewFile = Qt.QtCore.Signal(str)
    sigUpdate = Qt.QtCore.Signal()
    sigThreadFinished = Qt.QtCore.Signal()

    def __init__(self, file_lock, local_path, dirname,
                 sphere, arch, arch_ids, arches,
                 data_1d, data_2d,
                 parent=None):
        super().__init__(parent)
        self.local_path = local_path
        self.file_lock = file_lock
        self.dirname = dirname

        # Data Objects
        self.sphere = sphere
        self.arch = arch
        self.arch_ids = arch_ids
        self.arches = arches
        self.data_1d = data_1d
        self.data_2d = data_2d
        self.new_scan = True
        self.update_2d = True
        self.auto_last = True
        self.latest_idx = None
        self.new_scan_loaded = False
        # self.new_scan_name = None

        # Link UI
        self.ui = Ui_Form()
        self.ui.setupUi(self)

        # self.layout = self.ui.layout
        self.layout = self.ui.gridLayout
        self.defaultWidget = defaultWidget()
        self.defaultWidget.sigSetUserDefaults.connect(self.set_user_defaults)

        # Toolbar setup
        self.toolbar = QtWidgets.QToolBar('Tools')

        # File menu setup (part of toolbar)
        self.actionOpenFolder = QtWidgets.QAction()
        self.actionOpenFolder.setText('Open Folder')

        self.actionSetDefaults = QtWidgets.QAction()
        self.actionSetDefaults.setText('Advanced...')

        self.actionSaveDataAs = QtWidgets.QAction()
        self.actionSaveDataAs.setText('Save As')

        self.actionNewFile = QtWidgets.QAction()
        self.actionNewFile.setText('New')

        self.exportMenu = QtWidgets.QMenu()
        self.exportMenu.setTitle('Export')

        self.actionSaveImage = QtWidgets.QAction()
        self.actionSaveImage.setText('Current Image')
        self.exportMenu.addAction(self.actionSaveImage)

        self.actionSaveArray = QtWidgets.QAction()
        self.actionSaveArray.setText('Current 1D Array')
        self.exportMenu.addAction(self.actionSaveArray)
        
        self.paramMenu = QtWidgets.QMenu()
        self.paramMenu.setTitle('Config')
        
        self.actionSaveParams = QtWidgets.QAction()
        self.actionSaveParams.setText('Save')
        self.actionSaveParams.triggered.connect(self.defaultWidget.save_defaults)
        self.paramMenu.addAction(self.actionSaveParams)
        
        self.actionLoadParams = QtWidgets.QAction()
        self.actionLoadParams.setText('Load')
        self.actionLoadParams.triggered.connect(self.defaultWidget.load_defaults)
        self.paramMenu.addAction(self.actionLoadParams)
        
        self.paramMenu.addAction(self.actionSetDefaults)

        self.fileMenu = QtWidgets.QMenu()
        self.fileMenu.addAction(self.actionOpenFolder)
        self.fileMenu.addAction(self.actionNewFile)
        self.fileMenu.addAction(self.actionSaveDataAs)
        self.fileMenu.addMenu(self.exportMenu)

        self.fileButton = QtWidgets.QToolButton()
        self.fileButton.setText('File')
        self.fileButton.setPopupMode(QtWidgets.QToolButton.InstantPopup)
        self.fileButton.setMenu(self.fileMenu)
        
        self.paramButton = QtWidgets.QToolButton()
        self.paramButton.setText('Config')
        self.paramButton.setPopupMode(QtWidgets.QToolButton.InstantPopup)
        self.paramButton.setMenu(self.paramMenu)
        # End file menu setup

        self.toolbar.addWidget(self.fileButton)
        self.toolbar.addWidget(self.paramButton)
        # End toolbar setup

        self.layout.addWidget(self.toolbar, 0, 0, 1, 2)
        self.actionSetDefaults.triggered.connect(self.defaultWidget.show)
        
        self.ui.listScans.itemDoubleClicked.connect(self.scans_clicked)
        # self.ui.listScans.itemActivated.connect(self.scans_clicked)
        self.ui.listData.itemSelectionChanged.connect(self.data_changed)
        self.ui.listData.itemClicked.connect(self.data_changed)
        self.ui.show_all.clicked.connect(self.show_all)
        self.actionOpenFolder.triggered.connect(self.open_folder)
        self.actionSaveDataAs.triggered.connect(self.save_data_as)
        self.actionNewFile.triggered.connect(self.new_file)
        
        self.file_thread = fileHandlerThread(self.sphere, self.arch,
                                             self.file_lock,
                                             arch_ids=self.arch_ids,
                                             arches=self.arches,
                                             data_1d=self.data_1d,
                                             data_2d=self.data_2d)
        self.file_thread.sigTaskDone.connect(self.thread_finished)
        self.file_thread.sigNewFile.connect(self.sigNewFile.emit)
        self.file_thread.sigUpdate.connect(self.sigUpdate.emit)
        self.file_thread.start(Qt.QtCore.QThread.LowPriority)
        
    def load_starting_defaults(self):
        default_path = os.path.join(self.local_path, "last_defaults.json")
        if os.path.exists(default_path):
            self.defaultWidget.load_defaults(fname=default_path)
        else:
            self.defaultWidget.save_defaults(fname=default_path)

    def set_user_defaults(self):
        default_path = os.path.join(self.local_path, "last_defaults.json")
        self.defaultWidget.save_defaults(fname=default_path)

    def update(self):
        """Calls both update_scans and update_data.
        """
        # ic()
        # self.update_scans()
        self.update_data()
        
    def update_scans(self):
        """Takes in directory path and adds files in path to listScans
        """
        # ic()
        if not os.path.exists(self.dirname):
            return

        self.ui.listScans.clear()
        self.ui.listScans.addItem('..')

        for name in sorted(os.listdir(self.dirname)):
            abspath = os.path.join(self.dirname, name)
            if os.path.isdir(abspath):
                self.ui.listScans.addItem(name + '/')
            elif name.split('.')[-1] in ('h5', 'hdf5'):
                self.ui.listScans.addItem(name)
    
    def update_data(self):
        """Updates list with all arch ids.
        """
        # ic()
        if self.sphere.name == "null_main":
            return

        # with self.sphere.sphere_lock:
        _idxs = [str(i) for i in list(self.sphere.arches.index)]
        # ic(_idxs, self.data_1d.keys(), self.latest_idx, self.auto_last, self.new_scan_loaded)

        if len(_idxs) == 0:
            self.ui.listData.clear()
            # self.ui.listData.addItem('No Data')
            return

        lw = self.ui.listData
        items = [lw.item(x).text() for x in range(lw.count())]
        eq = _idxs == items
        # ic(_idxs, items, eq)

        if (len(_idxs) > 1) and (_idxs == items):
            if self.new_scan_loaded:
                self.new_scan_loaded = False
                self.ui.listData.setCurrentRow(-1)
                self.arch_ids = []
                return
            if self.auto_last and (self.latest_idx in _idxs) and (len(self.latest_idx) == 1):
                items = self.ui.listData.findItems(str(self.latest_idx), QtCore.Qt.MatchExactly)
                # ic(self.latest_idx, items)
                if len(items):
                    for item in items:
                        self.h5viewer.ui.listData.setCurrentItem(item)
                return
        if (len(_idxs) > 1) and (len(_idxs) == (len(items))):
            return

        previous_loc = self.ui.listData.currentRow()
        previous_sel = self.ui.listData.selectedItems()
        # self.ui.listData.itemSelectionChanged.disconnect(self.data_changed)

        # ic(previous_loc, previous_sel)

        self.ui.listData.clear()
        self.ui.listData.insertItems(0, _idxs)

        if self.new_scan_loaded:
            self.new_scan_loaded = False
            self.ui.listData.setCurrentRow(-1)
            self.arch_ids.clear()
            return

        # ic(self.auto_last, self.latest_idx, _idxs)
        if self.auto_last and isinstance(self.latest_idx, int) and (str(self.latest_idx) in _idxs):
            items = self.ui.listData.findItems(str(self.latest_idx), QtCore.Qt.MatchExactly)
            # ic(self.latest_idx, items)
            # self.ui.listData.itemSelectionChanged.connect(self.data_changed)
            if len(items):
                for item in items:
                    self.ui.listData.setCurrentItem(item)
            return
        # for idx in _idxs:
        #     self.ui.listData.addItem(str(idx))

        # if self.new_scan_name == self.sphere.name:

        # for idx in _idxs:
        #     self.ui.listData.addItem(str(idx))

        # if (len(_idxs) == 0) or (len(_idxs) > self.ui.listData.count() - 1):
        #     self.ui.listData.clear()
        #     # self.ui.listData.addItem('Overall')
        #     for idx in _idxs:
        #         self.ui.listData.addItem(str(idx))

        # lw = self.ui.listData
        # items = [lw.item(x).text() for x in range(lw.count())]

        if previous_loc > self.ui.listData.count() - 1:
            previous_loc = self.ui.listData.count() - 1

        # self.ui.listData.itemSelectionChanged.connect(self.data_changed)
        if len(previous_sel) < 2:
            self.ui.listData.setCurrentRow(previous_loc)
        else:
            for item in previous_sel:
                item.setSelected(True)

    def show_all(self):
        # ic()

        if len(self.sphere.arches.index) > 0:
            self.arch_ids.clear()
            self.arch_ids += self.sphere.arches.index

        # ic(self.arch_ids)
        self.new_scan = False
        self.data_changed(show_all=True)

    def thread_finished(self, task):
        # ic()
        if task != "load_arch":
            self.update()
        self.sigThreadFinished.emit()

        gc.collect()
    
    def scans_clicked(self, q):
        """Handles items being double clicked in listScans. Either
        navigates to new folder or loads in sphere data.
        
        q: QListItem, item selected in h5viewer.
        """
        # ic()
        try:
            if q.data(0) == '..':
                if self.dirname[-1] in ['/', '\\']:
                    up = os.path.dirname(self.dirname[:-1])
                else:
                    up = os.path.dirname(self.dirname)

                if os.path.isdir(up) and os.path.splitdrive(up)[1] != '':
                    self.dirname = up
                    self.update_scans()
            elif '/' in q.data(0):
                dirname = os.path.join(self.dirname, q.data(0))
                if os.path.isdir(dirname):
                    self.dirname = dirname
                    self.update_scans()
            elif q.data(0) != 'No scans':
                self.set_file(os.path.join(self.dirname, q.data(0)))
                self.new_scan_loaded = True
        except AttributeError:
            pass
    
    def set_file(self, fname):
        """Changes the data file.
        
        args:
            fname: str, absolute path for data file
        """
        # ic()
        if fname != '':
            try:
                # with self.file_lock:
                #     with catch_h5py_file(fname, 'a') as _:
                #         pass

                self.ui.listData.itemSelectionChanged.disconnect(self.data_changed)
                self.ui.listData.clear()
                self.ui.listData.addItem('Loading...')
                # self.set_open_enabled(False)
                self.file_thread.fname = fname
                self.file_thread.queue.put("set_datafile")
                self.ui.listData.itemSelectionChanged.connect(self.data_changed)
                self.new_scan = True
            except:
                traceback.print_exc()
                return

    def data_changed(self, show_all=False):
        """Connected to currentItemChanged signal of listData
        """
        # ic()
        if not show_all:
            self.arch_ids.clear()
            items = self.ui.listData.selectedItems()
            self.arch_ids += sorted([str(item.text()) for item in items])
            idxs = self.arch_ids
        else:
            idxs = self.arch_ids

        # ic(idxs, self.new_scan)

        if (len(idxs) == 0) or ('No data' in idxs):
            time.sleep(0.1)
            return

        # if idxs[0] == 'No data':
        #     self.arch_ids = []
        #     self.sigUpdate.emit()
        #     return

        # if self.new_scan and ('Overall' in idxs) and (len(self.data_1d) == 0):
        # if self.new_scan and ('Overall' in idxs):
        # if self.new_scan:
        #     self.new_scan = False
        #     return

        # ic(idxs, self.new_scan, len(self.data_1d), len(self.data_2d), len(self.sphere.arches.index))

        # Put 'Overall' first in list
        load_2d = self.update_2d
        # if 'Overall' in self.arch_ids:
        #     self.arch_ids.insert(0, self.arch_ids.pop(self.arch_ids.index('Overall')))
        #     idxs = self.sphere.arches.index

        if len(self.sphere.arches.index) > 1:
            if len(idxs) == len(self.sphere.arches.index):
                load_2d = False

        if load_2d:
            idxs_memory = [int(idx) for idx in idxs if int(idx) in self.data_2d.keys()]
        else:
            idxs_memory = [int(idx) for idx in idxs if int(idx) in self.data_1d.keys()]

        # ic(load_2d, idxs_memory)

        # Remove 2d data from 'Sum' if for unselected keys
        if load_2d:
            if (len(self.arches) == 0) or (len(self.data_2d) == 0):
                self.arches.update({'sum_int_2d': int_2d_data_static(), 'sum_map_raw': 0})
                self.arches.update({'idxs': [], 'add_idxs': [], 'sub_idxs': []})

            if len(idxs) > 1:
                self.get_arches_sum(idxs, idxs_memory)

            self.arches['idxs'] = [int(idx) for idx in idxs]

        # self.file_thread.arch_ids = [int(idx) for idx in idxs
        #                              if int(idx) not in idxs_memory]
        arch_ids = [int(idx) for idx in idxs
                    if int(idx) not in idxs_memory]

        # ic(arch_ids)
        # if len(self.file_thread.arch_ids) > 0:
        #     self.file_thread.update_2d = load_2d
        #     self.file_thread.queue.put("load_arches")
        if len(arch_ids) > 0:
            self.load_arches_data(arch_ids, load_2d)

        self.sigUpdate.emit()

    gc.collect()

    def data_reset(self):
        """Resets data in memory (self.arches, self.arch_ids, self.data_..
        """
        # ic()
        self.arches.clear()
        self.arch_ids.clear()
        self.data_1d.clear()
        self.data_2d.clear()
        self.new_scan = True

    def open_folder(self):
        """Changes the directory being displayed in the file explorer.
        """
        # ic()
        dirname = QFileDialog().getExistingDirectory(
            caption='Choose Directory',
            directory='',
            options=QFileDialog.ShowDirsOnly
        )
        if os.path.exists(dirname):
            self.dirname = dirname
            self.arches.clear()
            self.data_1d.clear()
            self.data_2d.clear()
            self.new_scan = True
            self.update_scans()
    
    def set_open_enabled(self, enable):
        """Sets the save and open actions to enable
        
        args:
            enable: bool, if True actions are enabled
        """
        # ic()
        self.actionSaveDataAs.setEnabled(enable)
        self.paramMenu.setEnabled(enable)
        self.actionOpenFolder.setEnabled(enable)
        self.actionNewFile.setEnabled(enable)
        # self.ui.listScans.setEnabled(enable)
    
    def save_data_as(self):
        """Saves all data to hdf5 file. Also sets fname to be the
        selected file.
        """
        # ic()
        fname, _ = QFileDialog.getSaveFileName()
        with self.file_thread.lock:
            self.file_thread.new_fname = fname
            self.file_thread.queue.put("save_data_as")
        self.set_file(fname)
    
    def new_file(self):
        """Calls file dialog and sets the file name.
        """
        # ic()
        fname, _ = QFileDialog.getSaveFileName()
        self.set_file(fname)

    def load_arches_data(self, arch_ids, load_2d):
        """Loads data from hdf5 file and sets attributes.

        args:
            file: h5py file or group object
        """
        # ic()
        with catch(self.sphere.data_file, 'r') as file:
            # ic(arch_ids)
            for idx in arch_ids:
                try:
                    # ic(idx)
                    arch = EwaldArch(idx=idx, static=True, gi=self.sphere.gi)
                    if not load_2d:
                        # arch.load_from_h5(file['arches'], load_2d=False)
                        self.load_arch_data(file['arches'], arch, idx, load_2d=False)
                        self.data_1d[int(idx)] = arch.copy(include_2d=False)
                        # ic('loaded 1D data', self.data_1d.keys())
                    else:
                        try:
                            if len(arch.int_2d.i_qChi) == 0:
                                pass
                        except TypeError:
                            arch.load_from_h5(file['arches'], load_2d=True)

                        self.data_1d[int(idx)] = arch.copy(include_2d=False)
                        self.data_2d[int(idx)] = {'map_raw': arch.map_raw,
                                                  'bg_raw': arch.bg_raw,
                                                  'mask': arch.mask,
                                                  'int_2d': arch.int_2d}

                        # ic('loaded 1 and 2D data', self.data_1d.keys(), self.data_2d.keys())
                        # ic(idx, self.arches['add_idxs'], self.arches['sub_idxs'])
                        if idx in self.arches['add_idxs']:
                            self.arches['sum_int_2d'] += self.data_2d[int(idx)]['int_2d']
                            # self.arches['sum_map_raw'] += self.data_2d[int(idx)]['map_raw']
                            self.arches['sum_map_raw'] += (self.data_2d[int(idx)]['map_raw'] -
                                                           self.data_2d[int(idx)]['bg_raw'])
                        elif idx in self.arches['sub_idxs']:
                            self.arches['sum_int_2d'] -= self.data_2d[int(idx)]['int_2d']
                            # self.arches['sum_map_raw'] -= self.data_2d[int(idx)]['map_raw']
                            self.arches['sum_map_raw'] -= (self.data_2d[int(idx)]['map_raw'] -
                                                           self.data_2d[int(idx)]['bg_raw'])

                except KeyError:
                    pass

    @staticmethod
    def load_arch_data(file, arch, idx, load_2d=True):
        # ic()
        if str(idx) not in file:
            print("No data can be found")
        else:
            grp = file[str(idx)]
            if 'type' in grp.attrs:
                if grp.attrs['type'] == 'EwaldArch':
                    if load_2d:
                        lst_attr = [
                            "map_raw", "mask", "map_norm", "scan_info", "ai_args",
                            "gi", "static", "poni_dict", "bg_raw"
                        ]
                        utils.h5_to_attributes(arch, grp, lst_attr)
                        arch.int_1d.from_hdf5(grp['int_1d'])
                        arch.int_2d.from_hdf5(grp['int_2d'])
                    else:
                        lst_attr = [
                            "scan_info", "ai_args",
                            "gi", "static", "poni_dict"
                        ]
                        utils.h5_to_attributes(arch, grp, lst_attr)
                        arch.int_1d.from_hdf5(grp['int_1d'])

                    # if self.poni_file is not None:
                    if arch.poni_dict is not None:
                        arch.integrator = create_ai_from_dict(arch.poni_dict)

    def get_arches_sum(self, idxs, idxs_memory):
        # ic()
        new_idxs = set([int(idx) for idx in idxs])
        old_idxs, data_keys = set(self.arches['idxs']), set(self.data_2d.keys())

        changed_idxs = new_idxs ^ old_idxs
        load_idxs = changed_idxs - data_keys

        # ic(new_idxs, old_idxs, changed_idxs, load_idxs)

        add_from_data = (new_idxs - old_idxs) & data_keys
        add_from_h5 = new_idxs - old_idxs - data_keys
        self.arches['add_idxs'] = [int(k) for k in add_from_h5]

        sub_from_data = (old_idxs - new_idxs) & data_keys
        sub_from_h5 = old_idxs - new_idxs - data_keys
        self.arches['sub_idxs'] = [int(k) for k in sub_from_h5]

        for x in load_idxs:
            if int(x) in idxs_memory:
                idxs_memory.remove(int(x))

        for k in add_from_data:
            try:
                self.arches['sum_int_2d'] += self.data_2d[int(k)]['int_2d']
                self.arches['sum_map_raw'] += self.data_2d[int(k)]['map_raw']
            except ValueError:
                self.arches['sum_int_2d'] = self.data_2d[int(k)]['int_2d']
                self.arches['sum_map_raw'] = self.data_2d[int(k)]['map_raw']

        for k in sub_from_data:
            try:
                self.arches['sum_int_2d'] -= self.data_2d[int(k)]['int_2d']
                self.arches['sum_map_raw'] -= self.data_2d[int(k)]['map_raw']
            except ValueError:
                self.arches['sum_int_2d'] = self.data_2d[int(k)]['int_2d']
                self.arches['sum_map_raw'] = self.data_2d[int(k)]['map_raw']



import socket
import time
import types
import sys
import yaml
import copy
import logging
from logging import info, warning, critical, debug
import asyncio


class BL_Variables():
    """Beamline variables are SPEC motors and counters """

    def __init__(self, bi):
        """Setup the BL_Variables object."""

        self.bi = bi

        pass

    class manage_beamline_variables():
        """This is basically a simple set of routines to create all of the motor variables and locate them in a simple object.
        I saw advantages to having both a dictionary and attributes so I am doing both (at least until i make up my mind). The 
        caller can ask for the full dictionary of motor_variaibles or access the motor_variable through an attribute set to the 
        motor_variable name
        
        Rather than having many small functions for each motor register, I have collected all of the necessary information in
        YAML config files and will create them on the fly. Some of the functions have more complex or custom motor interactions
        and those have been overloaded. Those overloaded functions are in motor_variaibles().

        """
        
        def __init__(self, var, bi):
            self.bi = bi

            self.var = var
            self._motor_var_list = []
            self._counter_var_list = []


            asyncio.ensure_future(self.setup_variables(), loop=self.bi.loop)


        def get_variables(self, include_hidden_variables = True):
            """Return the dictionary of motor variables. By default, it returns all variables, including the hidden
            ones. It can also return just the public list. Make more pythonic"""


            if include_hidden_variables is True:
                return self._motor_var_list, self._counter_var_list
            else: #filter the list by the hidden_from_client attribute
                ret_dict = {}


                #This is broken. Don't use
                for (motor_name, motor_var) in self._variables.items():
                    if motor_var.hidden_from_client is False:
                        ret_dict[motor_name] = motor_var

                return ret_dict


        def destroy_all_variables(self):
            """Kaboom! Delete all of the motor variables. This should cleanup/cancel all of the followers as well."""


            for beamline_var in self._counter_var_list:
                del beamline_var

            for beamline_var in self._motor_var_list:
                del beamline_var

        def __del__(self):
            self.destroy_all_variables()


        async def setup_variables(self):
            """Lets read the motors and counters from the beam line and create the variables"""

            #Start by reading the mnemonics
            await asyncio.sleep(1)
            motor_mnemonics = await self.bi.sis.get_all_motor_mnemonics()
            counter_mnemonics = await self.bi.sis.get_all_counter_mnemonics()


            #Loop through the motor mnemonics and create the motor variables
            motor_index = 0
            print("motor mnemonics", motor_mnemonics)
            for new_motor_name in motor_mnemonics:
                motor_var = BL_Variables.variable_object(self.bi, new_motor_name, None, is_motor=True, index_in_read=motor_index)
                print("{}: {}".format(new_motor_name, motor_var))
                setattr(self.var, new_motor_name, motor_var)
                self._motor_var_list.append(motor_var)
                motor_index = motor_index + 1


            #Loop through the counter mnemonics and create the
            counter_index = 0
            for new_counter_name in counter_mnemonics:
                counter_var = BL_Variables.variable_object(self.bi, new_counter_name, None, is_counter=True, index_in_read=counter_index)
                setattr(self.var, new_counter_name, counter_var)
                self._counter_var_list.append(counter_var)
                counter_index = counter_index + 1

            #Set the values for everything
            await self.refresh_variables()


        async def refresh_variables(self, update_counters=True, update_motors=True):
            """ Refresh all or some motor variables. a specific unit can be updated or all variables can be updated by 
            setting unit_to_update=None (default). Updates to certain motor variables may cause updates of others. 
            For example, enabling encoder or changing microsteps will cause variables with units='partial_step' to change 
            at the motor level. Thus, to prevent any significant divergence of the internal state, it can be useful to
            update the internal state to match that of the motor level (just in case we ever get careless and start
            directly accessing the internal cached values (e.g. bypassing the set_function and get_functions). Technically,
            this isn't necessary as any read request will trigger a read of the internal motor register. Still, it seems
            like a good idea to make sure the current cached value is not significantly out of date."""

            all_motor_positions = await self.bi.sis.get_all_motor_positions()
            all_motor_status = await self.bi.sis.get_all_motor_status()

            print("In refresh", self._motor_var_list)
            for motor_index in range(len(all_motor_positions)):
                motor_var = self._motor_var_list[motor_index]

                motor_value = all_motor_status[motor_index]
                motor_value['position'] = all_motor_positions[motor_index]

                motor_var.set_value(None, motor_value)

            all_counters = await self.bi.sis.get_all_counters()

            #Update all the counters
            for counter_index in range(len(all_counters)):
                counter_var = self._counter_var_list[counter_index]
                counter_value = all_counters[counter_index]

                counter_var.set_value(None, counter_value)
            pass



                        
                
    class variable_object():
        """Motor Variables give a place for high-level functions to attach followers (Python Futures) for specific command or oracle frame IDs. 
        When the frame updates the contents of a 'motor variaible', the result is sent to the followers (attached to this specific variable). 
        This function takes a 'description' of the motor_variaible that is being read from a YAML file by another routine ()"""
        def __init__(self, bi, variable_name, var_description, is_motor=None, is_counter=None, index_in_read = None):
            self.MOTOR ="motor"
            self.COUNTER = "counter"

            self.bi = bi #Keep the calling motor interactor as it could be handy
            self.beamline = bi.beamline #Handy to keep a motor object around
            self.followers = {}
            self.last_written_value = None
            self.last_command_id = None
            self.name = variable_name #This is the key used in the description
            self.params = {}

            self.var_type = self.MOTOR if is_motor is True else self.COUNTER

            self.last_changed = time.time()
            self.value = None

            self.index_in_read = index_in_read


            #I should try to load a config file associated w/ the beam line and check if this motor has any special stuff

            if is_motor is True:
                    self.set_function = None #Motor moves get executed in control frames!
                    self.get_function = self.get_motor
            elif is_counter is True:
                    self.set_function = None
                    self.get_function = self.get_counter

            #Create some useful shortcuts that I should probably be loading from a config file
            #self.typecast = self.params['typecast']
            #self.helptext = self.params['Help']
            #self.units = self.params['units']
            #self.hidden_from_client = self.params['hidden_from_client']


            
        def __del__(self):
            print("motor_variable '{}' being deleted. Sending cancel to all followers".format(self.name))
            #send a cancel to all of the futures on all frame IDs
            for frame_uuid in self.followers:
                self.cancel_followers(frame_uuid)
                
            del(self.followers)

        
        
            
        ########################################################################################################################
        # Manage the followers of this variable (add_follower is probably the only one in here anyone will need to call)
        ########################################################################################################################
        
        class return_to_follower():
            """Information to be returned"""
            def __init__(self, future, var):

                self.time_updated = time.time()
                self.value = copy.copy(var.value) #This should't be necessary but just in case the original object gets deleted, just give a copy.
                #self._units = copy.copy(var.units)
                self.future = future

                try:
                    #print("return_to_follower: setting result")
                    future.set_result(self)
                except CancelledError:
                    print("return_to_follower: Future {} was cancelled elsewhere.".format(future))
                except InvalidStateError:
                    print("return_to_follower: Future already done")
                except:
                    print("return_to_follower: bigger problem")
                    raise


        def update_followers(self, frame_uuid):
            """Copy the current result and send it to all of the followers with this specific cmd_uuid. The entire key is removed at the end."""
            #print(" in update")

            try:
                current_followers = self.get_followers(frame_uuid)
                for follower in current_followers:
                    #print("returning to follower")
                    self.return_to_follower(follower, self)
                
                #print("Updated {} followers".format(current_followers))
   
            except KeyError: #No Followers registered for this frame ID
                pass
            except TypeError: #frame_uuid = None or something not found
                #print("frame_uuid not found. No update")
                pass
            else: 
                del(self.followers[frame_uuid])



        def cancel_followers(self, frame_uuid):
            """Cancel all followers with a specific frame ID"""
            for follower in self.get_followers(frame_uuid):
                try:
                    follower.cancel()
                except:
                    pass


        def get_followers(self, frame_uuid):
            """Return the list of followers for the unique Control (or oracle) frame specified. If no followers, return 'None'"""
            try:
                return self.followers[frame_uuid]
            except KeyError:
                return None


        def add_follower(self, frame_uuid, future):
            """Add a follower to this variable. The unique Control (or Oracle) Frame ID needs to be supplied to ensure update happens when scheduled"""
            try: #Add to an existing frame_uuid
                self.followers[frame_uuid].append(future)
            except: #otherwise create the key and assign it to list containing the future
                self.followers[frame_uuid] = [future]

                
                
        ########################################################################################################################
        # Interactions with the motor variables go through the read_value and write_value functions.
        ########################################################################################################################
        async def read_value(self, frame_uuid):
            """Read the value of a variaible using the 'get_function' defined in the config file. The variable 'value'
            is updated and followers with specified frame_uuid are updated. Call with frame_uuid='None' to skip
            updating followers."""

            if self.get_function is not None:
                print("READ VALUE {}: {}".format(self.name, self.get_function))

                overloaded_fn = await self.get_function(frame_uuid)
                return overloaded_fn

            return None
            
        async def write_value(self, frame_uuid, value):
            """Write a value to the motor register using the 'set_function' defined in the config file. The 'get_function' is 
            called to update the 'value' and followers with specified frame_uuid are upddated. Call with frame_uuid='None' to 
            skip updating followers. Custom responses may be possible depending on implementation"""
            
            #Call the (possibly) overloaded function to write a value to a variable.
            if self.set_function is not None:
                await self.set_function(value, frame_uuid)
            
            #Trigger the read of the new value and return it to the user. This function will also trigger an update for the followers.

            return(await self.read_value(frame_uuid))
        
        def set_value(self, frame_uuid, value):
            """Set the value of the motor and update followers. Set frame_uuid to 'None' to avoid updating any followers. This function 
            should only appear when creating custom motor variable register read/write routines (e.g. read_many_vars or the custom_get_software_limits)"""
            
            
            self.value = value
            self.last_changed = time.time()

            #print(value)
            self.update_followers(frame_uuid)

        ########################################################################################################################
        # Default functions to set and get variable infromation directly from the motors and counters
        ########################################################################################################################


        async def get_motor(self, frame_uuid=None):
            """Get the position and status of a single motor"""


            motor_pos = await self.bi.sis.get_motor_position(self.name)
            motor_status = await self.bi.sis.get_motor_information(self.name)

            new_value = motor_status
            new_value['position'] = motor_pos

            #set the value and update the followers
            self.set_value(frame_uuid, new_value)

            return new_value

        async def get_counter(self, frame_uuid=None):
            """Get the value of a counter. Note: all the counters are read. While only one counter is returned, all are actually updated since we have the information"""

            all_counters = await self.bi.sis.get_all_counters()

            #Update all the counters
            for counter_index in range(len(all_counters)):
                counter_var = self.variables._counter_var_list[counter_index]
                counter_value = all_counters[counter_index]

                counter_var.set_value(frame_uuid, counter_value)

            return all_counters[self.index_to_read]







# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'imageWidgetUI.ui'
#
# Created by: PyQt5 UI code generator 5.15.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(1440, 842)
        self.horizontalLayout = QtWidgets.QHBoxLayout(Form)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setSpacing(0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.imageFrame = QtWidgets.QFrame(Form)
        self.imageFrame.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.imageFrame.setFrameShadow(QtWidgets.QFrame.Plain)
        self.imageFrame.setLineWidth(0)
        self.imageFrame.setObjectName("imageFrame")
        self.horizontalLayout.addWidget(self.imageFrame)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file '/Users/v/Research/RDA/repos/xdart/xdart/gui/static.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(800, 600)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.horizontalLayoutWidget = QtWidgets.QWidget(self.centralwidget)
        self.horizontalLayoutWidget.setGeometry(QtCore.QRect(40, 90, 160, 80))
        self.horizontalLayoutWidget.setObjectName("horizontalLayoutWidget")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.horizontalLayoutWidget)
        self.horizontalLayout.setSizeConstraint(QtWidgets.QLayout.SetMinimumSize)
        self.horizontalLayout.setContentsMargins(1, 1, 1, 1)
        self.horizontalLayout.setSpacing(2)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.toolButton = QtWidgets.QToolButton(self.horizontalLayoutWidget)
        self.toolButton.setObjectName("toolButton")
        self.horizontalLayout.addWidget(self.toolButton)
        self.pushButton = QtWidgets.QPushButton(self.horizontalLayoutWidget)
        self.pushButton.setObjectName("pushButton")
        self.horizontalLayout.addWidget(self.pushButton)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 800, 22))
        self.menubar.setObjectName("menubar")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "MainWindow"))
        self.toolButton.setText(_translate("MainWindow", "..."))
        self.pushButton.setText(_translate("MainWindow", "PushButton"))



import os

import pyFAI.resources
import pyFAI.calibrant
import pyFAI.detectors
import pyFAI.io.image
import fabio

from pyFAI.gui.CalibrationWindow import CalibrationWindow
from pyFAI.app.calib2 import parse_options, setup_model
from pyFAI.app.drawmask import MaskImageWidget
from pyFAI.gui.CalibrationContext import CalibrationContext

from pyqtgraph import Qt
import silx
from silx.gui import qt

import logging
logging.basicConfig(level=logging.INFO)
logging.captureWarnings(True)
logger = logging.getLogger(__name__)
try:
    import hdf5plugin  # noqa
except ImportError:
    logger.debug("Unable to load hdf5plugin, backtrace:", exc_info=True)

QFileDialog = Qt.QtWidgets.QFileDialog
logger_uncaught = logging.getLogger("pyFAI-calib2.UNCAUGHT")


def pyFAI_calib2_main():
    # It have to be done before loading Qt
    # --help must also work without Qt
    options = parse_options()

    if options.debug:
        logging.root.setLevel(logging.DEBUG)

    # Then we can load Qt
    if options.opengl:
        silx.config.DEFAULT_PLOT_BACKEND = "opengl"

    pyFAI.resources.silx_integration()
    settings = qt.QSettings(qt.QSettings.IniFormat,
                            qt.QSettings.UserScope,
                            "pyfai",
                            "pyfai-calib2",
                            None)

    context = CalibrationContext(settings)
    context.restoreSettings()

    setup_model(context.getCalibrationModel(), options)
    window = CalibrationWindowXdart(context)
    window.setVisible(True)
    window.setAttribute(qt.Qt.WA_DeleteOnClose, True)

    context.saveSettings()


# def pyFAI_drawmask_main(window, processFile):
def pyFAI_drawmask_main(window, image, processFile):
    usage = "pyFAI-drawmask file1.edf file2.edf ..."
    version = "pyFAI-average version %s from %s" % (pyFAI.version, pyFAI.date)
    description = """
    Draw a mask, i.e. an image containing the list of pixels which are considered invalid
    (no scintillator, module gap, beam stop shadow, ...).
    This will open a window and let you draw on the first image
    (provided) with different tools (brush, rectangle selection...)
    When you are finished, click on the "Save and quit" button.
    """
    epilog = """The mask image is saved into file1-masked.edf.
    Optionally the script will print the number of pixel masked
    and the intensity masked (as well on other files provided in input)"""

    # image = fabio.open(processFile).data
    window.setImageData(image)
    outfile = os.path.splitext(processFile)[0] + "-mask.edf"
    window.setOutputFile(outfile)
    window.outFile = outfile

    print("Your mask-file will be saved into %s" % outfile)


class CalibrationWindowXdart(CalibrationWindow):

    def __init__(self, context):
        super(CalibrationWindowXdart, self).__init__(context=context)
        self.context = context

    def closeEvent(self, event):
        poniFile = self.model().experimentSettingsModel().poniFile()

        if not poniFile.isSynchronized():
            button = qt.QMessageBox.question(
                self,
                "calib2",
                "The PONI file was not saved.\nDo you really want to close the application?",
                qt.QMessageBox.Cancel | qt.QMessageBox.No | qt.QMessageBox.Yes,
                qt.QMessageBox.Yes)
            if button != qt.QMessageBox.Yes:
                event.ignore()
                return

        event.accept()
        CalibrationContext._releaseSingleton()


class MaskImageWidgetXdart(MaskImageWidget):
    """
    Window application which allow to create a mask manually.
    It is based on Silx library widgets.
    """

    def __init__(self):
        super(MaskImageWidgetXdart, self).__init__()
        self.outFile = None

    def closeEvent(self, event):
        if os.path.exists(self.outFile):
            mask_file = os.path.basename(self.outFile)
            out_dialog = Qt.QtWidgets.QMessageBox()
            out_dialog.setText(f'{mask_file} saved in Image directory')
            out_dialog.exec_()
        event.accept()
        return



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'c:\Users\walroth\Documents\repos\xdart\xdart\experiments\ttheta_scan\gui\wranglers\specUI.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from pyqtgraph.Qt import QtCore, QtGui, QtWidgets

class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(400, 300)
        self.verticalLayout = QtWidgets.QVBoxLayout(Form)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.paramFrame = QtWidgets.QFrame(Form)
        self.paramFrame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.paramFrame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.paramFrame.setObjectName("paramFrame")
        self.verticalLayout.addWidget(self.paramFrame)
        self.specLabel = QtWidgets.QLabel(Form)
        self.specLabel.setText("")
        self.specLabel.setObjectName("specLabel")
        self.verticalLayout.addWidget(self.specLabel)
        self.frame = QtWidgets.QFrame(Form)
        self.frame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame.setObjectName("frame")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.frame)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.startButton = QtWidgets.QPushButton(self.frame)
        self.startButton.setObjectName("startButton")
        self.horizontalLayout.addWidget(self.startButton)
        self.pauseButton = QtWidgets.QPushButton(self.frame)
        self.pauseButton.setObjectName("pauseButton")
        self.horizontalLayout.addWidget(self.pauseButton)
        self.continueButton = QtWidgets.QPushButton(self.frame)
        self.continueButton.setObjectName("continueButton")
        self.horizontalLayout.addWidget(self.continueButton)
        self.stopButton = QtWidgets.QPushButton(self.frame)
        self.stopButton.setObjectName("stopButton")
        self.horizontalLayout.addWidget(self.stopButton)
        self.verticalLayout.addWidget(self.frame)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.startButton.setText(_translate("Form", "Start"))
        self.pauseButton.setText(_translate("Form", "Pause"))
        self.continueButton.setText(_translate("Form", "Continue"))
        self.stopButton.setText(_translate("Form", "Stop"))




from collections import OrderedDict
from pyFAI.detectors import Detector
import numpy as np

from xdart.utils.containers import PONI
from ..operation import Operation

inputs = OrderedDict(
    rotations = {
        "rot1": None,
        "rot2": None,
        "rot3": None
    },
    calib_rotations = {
        "rot1": 0,
        "rot2": 0,
        "rot3": 0
    },
    poni_file = None,
    spec_dict = {}
)

outputs = {
    'Distance': 0,
    'Poni1': 0,
    'Poni2': 0,
    'Rot1': 0,
    'Rot2': 0,
    'Rot3': 0,
    'Wavelength': 0,
    'Detector': 'Detector',
    'Detector_config': {
        'pixel1': 100e-6,
        'pixel2': 100e-6,
        'max_shape': None
    }
}

class MakePONI(Operation):
    """Operation for creating PONI objects. Uses a calibration poni file
    and adds in rotation values from motor positions to yield a
    dictionary with new values for a PONI object.
    """
    def __init__(self):
        super(MakePONI, self).__init__(inputs, outputs)
        self.base = None
    
    def run(self):
        """Main method of operation, calculates values for new PONI
        object.
        
        returns:
            outputs: dict, values to create a PONI object.
        """
        if self.base is None:
            self._set_base()
        poni = PONI.from_ponifile(self.inputs['poni_file'])
        
        for key, val in self.inputs['rotations'].items():
            if val is not None:
                r = np.radians(-self.inputs['spec_dict'][val]) + getattr(self.base, key)
                setattr(poni, key, r)
        
        self.outputs.update(poni.to_dict())

        return self.outputs
    
    def _set_base(self):
        """Adjusts base calibration based on where calibration was
        performed.
        """
        base = PONI.from_ponifile(self.inputs['poni_file'])
        for key, val in self.inputs['calib_rotations'].items():
            try:
                r = getattr(base, key) - val
            except TypeError:
                r = getattr(base, key)
            setattr(base, key, r)
        self.base = base

